# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""Layers manager for FreeCAD"""

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate

PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")


def getColorIcon(color):
    "returns a QtGui.QIcon from a color 3-float tuple"

    from PySide import QtGui

    c = QtGui.QColor(int(color[0] * 255), int(color[1] * 255), int(color[2] * 255))
    im = QtGui.QImage(48, 48, QtGui.QImage.Format_ARGB32)
    im.fill(c)
    px = QtGui.QPixmap.fromImage(im)
    return QtGui.QIcon(px)


class BIM_Layers:
    "The BIM_Layers FreeCAD command"

    def GetResources(self):

        return {
            "Pixmap": "BIM_Layers",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Layers", "Manage Layers"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "BIM_Layers", "Sets/modifies the different layers of your BIM project"
            ),
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):

        # only raise the dialog if it is already open
        if getattr(self, "dialog", None):
            self.dialog.raise_()
            return

        from PySide import QtGui

        # store changes to be committed
        self.deleteList = []

        # store assignlist
        self.assignList = {}

        # create the dialog
        self.dialog = FreeCADGui.PySideUic.loadUi(":/ui/dialogLayersIFC.ui")

        # store the ifc icon
        self.ifcicon = QtGui.QIcon(":/icons/IFC.svg")

        # set nice icons
        self.dialog.setWindowIcon(QtGui.QIcon(":/icons/Draft_Layer.svg"))
        self.dialog.buttonNew.setIcon(QtGui.QIcon(":/icons/document-new.svg"))
        self.dialog.buttonDelete.setIcon(QtGui.QIcon(":/icons/delete.svg"))
        self.dialog.buttonSelectAll.setIcon(QtGui.QIcon(":/icons/edit-select-all.svg"))
        self.dialog.buttonToggle.setIcon(QtGui.QIcon(":/icons/dagViewVisible.svg"))
        self.dialog.buttonIsolate.setIcon(QtGui.QIcon(":/icons/Std_ShowSelection.svg"))
        self.dialog.buttonCancel.setIcon(QtGui.QIcon(":/icons/edit_Cancel.svg"))
        self.dialog.buttonOK.setIcon(QtGui.QIcon(":/icons/edit_OK.svg"))
        self.dialog.buttonAssign.setIcon(QtGui.QIcon(":/icons/button_right.svg"))
        self.dialog.buttonIFC.setIcon(self.ifcicon)

        # restore window geometry from stored state
        w = PARAMS.GetInt("LayersManagerWidth", 640)
        h = PARAMS.GetInt("LayersManagerHeight", 320)
        self.dialog.resize(w, h)

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.dialog.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.dialog.rect().center()
        )

        # connect signals/slots
        self.dialog.buttonNew.clicked.connect(self.addItem)
        self.dialog.buttonDelete.clicked.connect(self.onDelete)
        self.dialog.buttonSelectAll.clicked.connect(self.dialog.tree.selectAll)
        self.dialog.buttonToggle.clicked.connect(self.onToggle)
        self.dialog.buttonCancel.clicked.connect(self.dialog.reject)
        self.dialog.buttonIsolate.clicked.connect(self.onIsolate)
        self.dialog.buttonOK.clicked.connect(self.accept)
        self.dialog.buttonAssign.clicked.connect(self.onAssign)
        self.dialog.buttonIFC.clicked.connect(self.onIFC)
        self.dialog.rejected.connect(self.reject)

        # set the model up
        self.model = QtGui.QStandardItemModel()
        self.dialog.tree.setModel(self.model)
        self.dialog.tree.setUniformRowHeights(True)
        self.dialog.tree.setItemDelegate(BIM_Layers_Delegate())
        self.dialog.tree.setItemsExpandable(False)
        self.dialog.tree.setRootIsDecorated(False)  # removes spacing in first column
        self.dialog.tree.setSelectionMode(
            QtGui.QTreeView.ExtendedSelection
        )  # allow one to select many

        # fill the tree view
        self.update()

        # rock 'n roll!!!
        self.dialog.show()

    def accept(self):
        "when OK button is pressed"

        import Draft

        doc = FreeCAD.ActiveDocument
        changed = False

        # delete layers
        for name in self.deleteList:
            if not changed:
                doc.openTransaction("Layers change")
                changed = True
            doc.removeObject(name)

        # assign
        for target, objs in self.assignList.items():
            target_obj = doc.getObject(target)
            if target_obj:
                for obj in objs:
                    target_obj.Proxy.addObject(target_obj, obj)

        # apply changes
        for row in range(self.model.rowCount()):
            # get or create layer
            name = self.model.item(row, 1).toolTip()
            obj = None
            if name:
                obj = doc.getObject(name)
            if not obj:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                if self.model.item(row, 1).icon().isNull():
                    obj = Draft.make_layer(self.model.item(row, 1).text())
                    # By default BIM layers should not swallow their children otherwise
                    # they will disappear from the tree root
                    obj.ViewObject.addProperty(
                        "App::PropertyBool", "HideChildren", "Layer", locked=True
                    )
                    obj.ViewObject.HideChildren = True
                else:
                    from nativeifc import ifc_tools
                    import FreeCADGui

                    active = FreeCADGui.ActiveDocument.ActiveView.getActiveObject("NativeIFC")
                    project = None
                    if active:
                        project = ifc_tools.get_project(active)
                    else:
                        projects = [
                            o
                            for o in doc.Objects
                            if hasattr(o, "Proxy") and hasattr(o.Proxy, "ifcfile")
                        ]
                        if projects:
                            project = projects[0]
                            if len(projects) > 1:
                                FreeCAD.Console.PrintWarning(
                                    translate(
                                        "BIM",
                                        "Warning: The new layer was added to the project",
                                    )
                                    + " "
                                    + project.Label
                                    + "\n"
                                )
                        else:
                            FreeCAD.Console.PrintError(
                                translate("BIM", "There is no IFC project in this document") + "\n"
                            )
                    if project:
                        obj = ifc_tools.create_layer(self.model.item(row, 1).text(), project)
            vobj = obj.ViewObject

            # visibility
            checked = self.model.item(row, 0).checkState() == QtCore.Qt.Checked
            if checked != vobj.Visibility:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.Visibility = checked

            # label
            label = self.model.item(row, 1).text()
            # Setting Label="" is possible in the Property editor but we avoid it here:
            if label and obj.Label != label:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                obj.Label = label

            # line width
            width = self.model.item(row, 2).data(QtCore.Qt.DisplayRole)
            # Setting LineWidth=0 is possible in the Property editor but we avoid it here:
            if width and vobj.LineWidth != width:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.LineWidth = width

            # draw style
            style = self.model.item(row, 3).text()
            if style is not None and vobj.DrawStyle != style:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.DrawStyle = style

            # line color
            color = self.model.item(row, 4).data(QtCore.Qt.UserRole)
            if color is not None and vobj.LineColor[3:] != color:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.LineColor = color

            # shape color
            color = self.model.item(row, 5).data(QtCore.Qt.UserRole)
            if color is not None and vobj.ShapeColor[3:] != color:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.ShapeColor = color

            # transparency
            transparency = self.model.item(row, 6).data(QtCore.Qt.DisplayRole)
            if transparency is not None and vobj.Transparency != transparency:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.Transparency = transparency

            # line print color
            color = self.model.item(row, 7).data(QtCore.Qt.UserRole)
            if color is not None and vobj.LinePrintColor[3:] != color:
                if not changed:
                    doc.openTransaction("Layers change")
                    changed = True
                vobj.LinePrintColor = color

        # recompute
        if changed:
            doc.commitTransaction()
            doc.recompute()

        # exit
        return self.dialog.reject()

    def reject(self):
        "when Cancel button is pressed or dialog is closed"

        # save dialog size
        pref = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM")
        pref.SetInt("LayersManagerWidth", self.dialog.width())
        pref.SetInt("LayersManagerHeight", self.dialog.height())

        # wipe to let FreeCAD know the dialog has been closed
        del self.dialog

        return True

    def update(self):
        "rebuild the model from document contents"

        import Draft

        self.model.clear()

        # set header
        self.model.setHorizontalHeaderLabels(
            [
                translate("BIM", "On"),
                translate("BIM", "Name"),
                translate("BIM", "Line width"),
                translate("BIM", "Draw style"),
                translate("BIM", "Line color"),
                translate("BIM", "Face color"),
                translate("BIM", "Transparency"),
                translate("BIM", "Line print color"),
            ]
        )
        self.dialog.tree.header().setDefaultSectionSize(72)
        self.dialog.tree.setColumnWidth(0, 32)  # on/off column
        self.dialog.tree.setColumnWidth(1, 128)  # name column

        # populate
        objs = [obj for obj in FreeCAD.ActiveDocument.Objects if Draft.getType(obj) == "Layer"]
        objs.sort(key=lambda o: o.Label)
        for obj in objs:
            self.addItem(obj)

    def addItem(self, obj=None):
        "adds a row to the model"

        from PySide import QtCore, QtGui

        # create row with default values
        onItem = QtGui.QStandardItem()
        onItem.setCheckable(True)
        onItem.setCheckState(QtCore.Qt.Checked)
        nameItem = QtGui.QStandardItem(translate("BIM", "New Layer"))
        widthItem = QtGui.QStandardItem()
        widthItem.setData(
            self.getPref("DefaultShapeLineWidth", 2, "Integer"), QtCore.Qt.DisplayRole
        )
        styleItem = QtGui.QStandardItem("Solid")
        lineColorItem = QtGui.QStandardItem()
        lineColorItem.setData(self.getPref("DefaultShapeLineColor", 421075455), QtCore.Qt.UserRole)
        shapeColorItem = QtGui.QStandardItem()
        shapeColorItem.setData(self.getPref("DefaultShapeColor", 3435973887), QtCore.Qt.UserRole)
        transparencyItem = QtGui.QStandardItem()
        transparencyItem.setData(0, QtCore.Qt.DisplayRole)
        linePrintColorItem = QtGui.QStandardItem()
        linePrintColorItem.setData(self.getPref("DefaultPrintColor", 0), QtCore.Qt.UserRole)
        if FreeCADGui.ActiveDocument.ActiveView.getActiveObject("NativeIFC"):
            nameItem.setIcon(self.ifcicon)

        # populate with object data
        if obj:
            if hasattr(obj, "StepId"):
                nameItem.setIcon(self.ifcicon)
            onItem.setCheckState(
                QtCore.Qt.Checked if obj.ViewObject.Visibility else QtCore.Qt.Unchecked
            )
            nameItem.setText(obj.Label)
            nameItem.setToolTip(obj.Name)
            widthItem.setData(obj.ViewObject.LineWidth, QtCore.Qt.DisplayRole)
            styleItem.setText(obj.ViewObject.DrawStyle)
            lineColorItem.setData(obj.ViewObject.LineColor[:3], QtCore.Qt.UserRole)
            shapeColorItem.setData(obj.ViewObject.ShapeColor[:3], QtCore.Qt.UserRole)
            transparencyItem.setData(obj.ViewObject.Transparency, QtCore.Qt.DisplayRole)
            if hasattr(obj.ViewObject, "LinePrintColor"):
                linePrintColorItem.setData(obj.ViewObject.LinePrintColor[:3], QtCore.Qt.UserRole)
        lineColorItem.setIcon(getColorIcon(lineColorItem.data(QtCore.Qt.UserRole)))
        shapeColorItem.setIcon(getColorIcon(shapeColorItem.data(QtCore.Qt.UserRole)))
        linePrintColorItem.setIcon(getColorIcon(linePrintColorItem.data(QtCore.Qt.UserRole)))

        # append row
        self.model.appendRow(
            [
                onItem,
                nameItem,
                widthItem,
                styleItem,
                lineColorItem,
                shapeColorItem,
                transparencyItem,
                linePrintColorItem,
            ]
        )

    def getPref(self, value, default, valuetype="Unsigned"):
        "retrieves a view pref value"

        p = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/View")
        if valuetype == "Unsigned":
            c = p.GetUnsigned(value, default)
            r = float((c >> 24) & 0xFF) / 255.0
            g = float((c >> 16) & 0xFF) / 255.0
            b = float((c >> 8) & 0xFF) / 255.0
            return (
                r,
                g,
                b,
            )
        elif valuetype == "Integer":
            return p.GetInt(value, default)

    def onDelete(self):
        "delete selected rows"

        rows = []
        for index in self.dialog.tree.selectedIndexes():
            if not index.row() in rows:
                rows.append(index.row())

            # append layer name to the delete list
            if index.column() == 1:
                name = self.model.itemFromIndex(index).toolTip()
                if name:
                    if not name in self.deleteList:
                        self.deleteList.append(name)

        # delete rows starting from the lowest, to not alter row indexes while deleting
        rows.sort()
        rows.reverse()
        for row in rows:
            self.model.takeRow(row)

    def onToggle(self):
        "toggle selected layers on/off"

        from PySide import QtCore

        state = None
        for index in self.dialog.tree.selectedIndexes():
            if index.column() == 0:
                # get state from first selected row
                if state is None:
                    if self.model.itemFromIndex(index).checkState() == QtCore.Qt.Checked:
                        state = QtCore.Qt.Unchecked
                    else:
                        state = QtCore.Qt.Checked
                self.model.itemFromIndex(index).setCheckState(state)

    def onIsolate(self):
        "isolates the selected layers (turns all the others off"

        from PySide import QtCore

        onrows = []
        for index in self.dialog.tree.selectedIndexes():
            if not index.row() in onrows:
                onrows.append(index.row())
        for row in range(self.model.rowCount()):
            if not row in onrows:
                self.model.item(row, 0).setCheckState(QtCore.Qt.Unchecked)

    def onIFC(self):
        "attributes this layer to an IFC project"

        from PySide import QtGui

        for index in self.dialog.tree.selectedIndexes():
            if index.column() == 1:
                name = self.model.itemFromIndex(index).toolTip()
                if name:
                    obj = FreeCAD.ActiveDocument.getObject(name)
                    if obj:
                        if hasattr(obj, "StepId"):
                            return
                item = self.model.itemFromIndex(index)
                if item.icon().isNull():
                    item.setIcon(self.ifcicon)
                else:
                    item.setIcon(QtGui.QIcon())

    def onAssign(self):
        "attributes selected objects to a selected layer"

        for index in self.dialog.tree.selectedIndexes():
            if index.column() == 1:
                name = self.model.itemFromIndex(index).toolTip()
                if name:
                    target = FreeCAD.ActiveDocument.getObject(name)
                    if target:
                        selected = FreeCADGui.Selection.getSelection()
                        if selected:
                            self.assignList.setdefault(target.Name, [])
                            for i in selected:
                                if i not in self.assignList[target.Name]:
                                    self.assignList[target.Name].append(i)


if FreeCAD.GuiUp:
    from PySide import QtCore, QtGui

    class BIM_Layers_Delegate(QtGui.QStyledItemDelegate):
        "model delegate"

        def __init__(self, parent=None, *args):
            QtGui.QStyledItemDelegate.__init__(self, parent, *args)
            # setEditorData() is triggered several times.
            # But we want to show the color dialog only the first time
            self.first = True

        def createEditor(self, parent, option, index):
            if index.column() == 0:  # Layer on/off
                editor = QtGui.QCheckBox(parent)
            if index.column() == 1:  # Layer name
                editor = QtGui.QLineEdit(parent)
            elif index.column() == 2:  # Line width
                editor = QtGui.QSpinBox(parent)
                editor.setMaximum(99)
            elif index.column() == 3:  # Line style
                editor = QtGui.QComboBox(parent)
                editor.addItems(["Solid", "Dashed", "Dotted", "Dashdot"])
            elif index.column() == 4:  # Line color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            elif index.column() == 5:  # Shape color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            elif index.column() == 6:  # Transparency
                editor = QtGui.QSpinBox(parent)
                editor.setMaximum(100)
            elif index.column() == 7:  # Line print color
                editor = QtGui.QLineEdit(parent)
                self.first = True
            return editor

        def setEditorData(self, editor, index):
            if index.column() == 0:  # Layer on/off
                editor.setChecked(index.data())
            elif index.column() == 1:  # Layer name
                editor.setText(index.data())
            elif index.column() == 2:  # Line width
                editor.setValue(index.data())
            elif index.column() == 3:  # Line style
                editor.setCurrentIndex(["Solid", "Dashed", "Dotted", "Dashdot"].index(index.data()))
            elif index.column() == 4:  # Line color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(
                        QtGui.QColor(int(c[0] * 255), int(c[1] * 255), int(c[2] * 255))
                    )
                    editor.setText(str(color.getRgbF()))
                    self.first = False
            elif index.column() == 5:  # Shape color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(
                        QtGui.QColor(int(c[0] * 255), int(c[1] * 255), int(c[2] * 255))
                    )
                    editor.setText(str(color.getRgbF()))
                    self.first = False
            elif index.column() == 6:  # Transparency
                editor.setValue(index.data())
            elif index.column() == 7:  # Line print color
                editor.setText(str(index.data(QtCore.Qt.UserRole)))
                if self.first:
                    c = index.data(QtCore.Qt.UserRole)
                    color = QtGui.QColorDialog.getColor(
                        QtGui.QColor(int(c[0] * 255), int(c[1] * 255), int(c[2] * 255))
                    )
                    editor.setText(str(color.getRgbF()))
                    self.first = False

        def setModelData(self, editor, model, index):
            if index.column() == 0:  # Layer on/off
                model.setData(index, editor.isChecked())
            elif index.column() == 1:  # Layer name
                model.setData(index, editor.text())
            elif index.column() == 2:  # Line width
                model.setData(index, editor.value())
            elif index.column() == 3:  # Line style
                model.setData(
                    index,
                    ["Solid", "Dashed", "Dotted", "Dashdot"][editor.currentIndex()],
                )
            elif index.column() == 4:  # Line color
                model.setData(index, eval(editor.text()), QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))
            elif index.column() == 5:  # Shape color
                model.setData(index, eval(editor.text()), QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))
            elif index.column() == 6:  # Transparency
                model.setData(index, editor.value())
            elif index.column() == 7:  # Line prin color
                model.setData(index, eval(editor.text()), QtCore.Qt.UserRole)
                model.itemFromIndex(index).setIcon(getColorIcon(eval(editor.text())))


FreeCADGui.addCommand("BIM_Layers", BIM_Layers())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *   Copyright (c) 2018, 2019 Gauthier Briere                              *
# *   Copyright (c) 2019, 2020 Schildkroet                                  *
# *   Copyright (c) 2022-2025 Larry Woestman <LarryWoestman2@gmail.com>     *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
These are functions related to arguments and values for creating custom post processors.
"""
import argparse
import os
import shlex
from typing import Any, Callable, Dict, Tuple, Union

from FreeCAD import Units

import Path.Post.UtilsParse as PostUtilsParse

# Define some types that are used throughout this file
PathParameter = float
PathParameters = Dict[str, PathParameter]
Parser = argparse.ArgumentParser
Values = Dict[str, Any]

ParameterFunction = Callable[[Values, str, str, PathParameter, PathParameters], str]


def add_flag_type_arguments(
    argument_group,
    default_flag: bool,
    true_argument: str,
    false_argument: str,
    true_help: str,
    false_help: str,
    visible: bool = True,
) -> None:
    """Create an argument specification for an argument that is a flag."""
    if visible:
        if default_flag:
            true_help += " (default)"
        else:
            false_help += " (default)"
    else:
        true_help = false_help = argparse.SUPPRESS
    argument_group.add_argument(true_argument, action="store_true", help=true_help)
    argument_group.add_argument(false_argument, action="store_true", help=false_help)


def init_argument_defaults(argument_defaults: Dict[str, bool]) -> None:
    """Initialize which argument to show as the default in flag-type arguments."""
    argument_defaults["axis-modal"] = False
    argument_defaults["bcnc"] = False
    argument_defaults["comments"] = True
    argument_defaults["enable_coolant"] = False
    argument_defaults["enable_machine_specific_commands"] = False
    argument_defaults["header"] = True
    argument_defaults["line-numbers"] = False
    argument_defaults["list_tools_in_preamble"] = False
    argument_defaults["metric_inches"] = True
    argument_defaults["modal"] = False
    argument_defaults["output_all_arguments"] = False
    argument_defaults["output_machine_name"] = False
    argument_defaults["output_path_labels"] = False
    argument_defaults["output_visible_arguments"] = False
    argument_defaults["show-editor"] = True
    argument_defaults["tlo"] = True
    argument_defaults["tool_change"] = True
    argument_defaults["translate_drill"] = False


def init_arguments_visible(arguments_visible: Dict[str, bool]) -> None:
    """Initialize the flags for which arguments are visible in the arguments tooltip."""
    arguments_visible["axis-modal"] = True
    arguments_visible["axis-precision"] = True
    arguments_visible["bcnc"] = False
    arguments_visible["chipbreaking_amount"] = False
    arguments_visible["command_space"] = False
    arguments_visible["comments"] = True
    arguments_visible["comment_symbol"] = False
    arguments_visible["enable_coolant"] = False
    arguments_visible["enable_machine_specific_commands"] = False
    arguments_visible["end_of_line_characters"] = False
    arguments_visible["feed-precision"] = True
    arguments_visible["finish_label"] = False
    arguments_visible["header"] = True
    arguments_visible["line_number_increment"] = False
    arguments_visible["line_number_start"] = False
    arguments_visible["line-numbers"] = True
    arguments_visible["list_tools_in_preamble"] = False
    arguments_visible["metric_inches"] = True
    arguments_visible["modal"] = True
    arguments_visible["output_all_arguments"] = True
    arguments_visible["output_machine_name"] = False
    arguments_visible["output_path_labels"] = False
    arguments_visible["output_visible_arguments"] = True
    arguments_visible["postamble"] = True
    arguments_visible["post_operation"] = False
    arguments_visible["preamble"] = True
    arguments_visible["precision"] = True
    arguments_visible["return-to"] = False
    arguments_visible["show-editor"] = True
    arguments_visible["tlo"] = True
    arguments_visible["tool_change"] = False
    arguments_visible["translate_drill"] = False
    arguments_visible["wait-for-spindle"] = False


def init_shared_arguments(
    values: Values,
    argument_defaults: Dict[str, bool],
    arguments_visible: Dict[str, bool],
) -> Parser:
    """Initialize the arguments for postprocessors."""
    help_message: str
    parser: Parser
    shared: argparse._ArgumentGroup

    parser = argparse.ArgumentParser(
        prog=values["MACHINE_NAME"], usage=argparse.SUPPRESS, add_help=False
    )
    shared = parser.add_argument_group("Arguments that are commonly used")
    add_flag_type_arguments(
        shared,
        argument_defaults["metric_inches"],
        "--metric",
        "--inches",
        "Convert output for Metric mode (G21)",
        "Convert output for US imperial mode (G20)",
        arguments_visible["metric_inches"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["axis-modal"],
        "--axis-modal",
        "--no-axis-modal",
        "Don't output axis values if they are the same as the previous line",
        "Output axis values even if they are the same as the previous line",
        arguments_visible["axis-modal"],
    )
    if arguments_visible["axis-precision"]:
        help_message = (
            f"Number of digits of precision for axis moves, "
            f'default is {str(values["DEFAULT_AXIS_PRECISION"])}'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--axis-precision",
        default=-1,
        type=int,
        help=help_message,
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["bcnc"],
        "--bcnc",
        "--no-bcnc",
        (
            "Add Job operations as bCNC block headers. "
            "Consider suppressing comments by adding --no-comments"
        ),
        "Suppress bCNC block header output",
        arguments_visible["bcnc"],
    )
    if arguments_visible["chipbreaking_amount"]:
        help_message = (
            "Amount to move for chipbreaking in a translated G73 command, "
            f'default is {str(values["CHIPBREAKING_AMOUNT"])}'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--chipbreaking_amount",
        help=help_message,
    )
    if arguments_visible["command_space"]:
        help_message = (
            "The character to use between parts of a command, "
            "default is a space, may also use a null string"
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--command_space",
        default=" ",
        help=help_message,
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["comments"],
        "--comments",
        "--no-comments",
        "Output comments",
        "Suppress comment output",
        arguments_visible["comments"],
    )
    if arguments_visible["comment_symbol"]:
        help_message = (
            f'The character used to start a comment, default is "{values["COMMENT_SYMBOL"]}"'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--comment_symbol",
        help=help_message,
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["enable_coolant"],
        "--enable_coolant",
        "--disable_coolant",
        "Enable coolant",
        "Disable coolant",
        arguments_visible["enable_coolant"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["enable_machine_specific_commands"],
        "--enable_machine_specific_commands",
        "--disable_machine_specific_commands",
        "Enable machine specific commands of the form (MC_RUN_COMMAND: blah)",
        "Disable machine specific commands",
        arguments_visible["enable_machine_specific_commands"],
    )
    if arguments_visible["end_of_line_characters"]:
        help_message = (
            "The character(s) to use at the end of each line in the output file, "
            "default is whatever the system uses, may also use '\\n', '\\r', or '\\r\\n'"
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--end_of_line_characters",
        help=help_message,
    )
    if arguments_visible["feed-precision"]:
        help_message = (
            f"Number of digits of precision for feed rate, "
            f'default is {str(values["DEFAULT_FEED_PRECISION"])}'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--feed-precision",
        default=-1,
        type=int,
        help=help_message,
    )
    if arguments_visible["finish_label"]:
        help_message = (
            "The characters to use in the 'Finish operation' comment, "
            f'default is "{values["FINISH_LABEL"]}"'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--finish_label",
        help=help_message,
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["header"],
        "--header",
        "--no-header",
        "Output headers",
        "Suppress header output",
        arguments_visible["header"],
    )
    if arguments_visible["line_number_increment"]:
        help_message = (
            f'Amount to increment the line numbers, default is {str(values["LINE_INCREMENT"])}'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--line_number_increment",
        default=-1,
        type=int,
        help=help_message,
    )
    if arguments_visible["line_number_start"]:
        help_message = (
            f'The number the line numbers start at, default is {str(values["line_number"])}'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--line_number_start",
        default=-1,
        type=int,
        help=help_message,
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["line-numbers"],
        "--line-numbers",
        "--no-line-numbers",
        "Prefix with line numbers",
        "Don't prefix with line numbers",
        arguments_visible["line-numbers"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["list_tools_in_preamble"],
        "--list_tools_in_preamble",
        "--no-list_tools_in_preamble",
        "List the tools used in the operation in the preamble",
        "Don't list the tools used in the operation",
        arguments_visible["list_tools_in_preamble"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["modal"],
        "--modal",
        "--no-modal",
        "Don't output the G-command name if it is the same as the previous line",
        "Output the G-command name even if it is the same as the previous line",
        arguments_visible["modal"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["output_all_arguments"],
        "--output_all_arguments",
        "--no-output_all_arguments",
        "Output all of the available arguments",
        "Don't output all of the available arguments",
        arguments_visible["output_all_arguments"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["output_machine_name"],
        "--output_machine_name",
        "--no-output_machine_name",
        "Output the machine name in the pre-operation information",
        "Don't output the machine name in the pre-operation information",
        arguments_visible["output_machine_name"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["output_path_labels"],
        "--output_path_labels",
        "--no-output_path_labels",
        "Output Path labels at the beginning of each Path",
        "Don't output Path labels at the beginning of each Path",
        arguments_visible["output_path_labels"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["output_visible_arguments"],
        "--output_visible_arguments",
        "--no-output_visible_arguments",
        "Output all of the visible arguments",
        "Don't output the visible arguments",
        arguments_visible["output_visible_arguments"],
    )
    if arguments_visible["postamble"]:
        help_message = (
            f"Set commands to be issued after the last command, "
            f'default is "{values["POSTAMBLE"]}"'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument("--postamble", help=help_message)
    if arguments_visible["post_operation"]:
        help_message = (
            f"Set commands to be issued after every operation, "
            f'default is "{values["POST_OPERATION"]}"'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument("--post_operation", help=help_message)
    if arguments_visible["preamble"]:
        help_message = (
            f"Set commands to be issued before the first command, "
            f'default is "{values["PREAMBLE"]}"'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument("--preamble", help=help_message)
    # The --precision argument is included for backwards compatibility with some
    # postprocessors.  If both --axis-precision and --precision are provided, the
    # --axis-precision value "wins".  If both --feed-precision and --precision are
    # provided, the --feed-precision value "wins".
    if arguments_visible["precision"]:
        help_message = (
            f"Number of digits of precision for both feed rate and axis "
            f'moves, default is {str(values["DEFAULT_AXIS_PRECISION"])} for metric or '
            f'{str(values["DEFAULT_INCH_AXIS_PRECISION"])} for inches'
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument(
        "--precision",
        default=-1,
        type=int,
        help=help_message,
    )
    if arguments_visible["return-to"]:
        help_message = (
            "Move to the specified x,y,z coordinates at the end, e.g. "
            "--return-to=0,0,0 (default is do not move)"
        )
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument("--return-to", default="", help=help_message)
    add_flag_type_arguments(
        shared,
        argument_defaults["show-editor"],
        "--show-editor",
        "--no-show-editor",
        "Pop up editor before writing output",
        "Don't pop up editor before writing output",
        arguments_visible["show-editor"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["tlo"],
        "--tlo",
        "--no-tlo",
        "Output tool length offset (G43) following tool changes",
        "Suppress tool length offset (G43) following tool changes",
        arguments_visible["tlo"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["tool_change"],
        "--tool_change",
        "--no-tool_change",
        "Insert M6 and any other tool change G-code for all tool changes",
        "Convert M6 to a comment for all tool changes",
        arguments_visible["tool_change"],
    )
    add_flag_type_arguments(
        shared,
        argument_defaults["translate_drill"],
        "--translate_drill",
        "--no-translate_drill",
        "Translate drill cycles G73, G81, G82 & G83 into G0/G1 movements",
        "Don't translate drill cycles G73, G81, G82 & G83 into G0/G1 movements",
        arguments_visible["translate_drill"],
    )
    if arguments_visible["wait-for-spindle"]:
        help_message = "Time to wait (in seconds) after M3, M4 (default = 0.0)"
    else:
        help_message = argparse.SUPPRESS
    shared.add_argument("--wait-for-spindle", type=float, default=0.0, help=help_message)
    return parser


def init_shared_values(values: Values) -> None:
    """Initialize the default values in postprocessors."""
    #
    # The starting axis precision is 3 digits after the decimal point.
    #
    values["AXIS_PRECISION"] = 3
    #
    # How far to move up (in millimeters) in the Z axis when chipbreaking
    # with a G73 command.
    #
    values["CHIPBREAKING_AMOUNT"] = Units.Quantity(0.25, Units.Length)

    #
    # If this is set to "", all spaces are removed from between commands and parameters.
    #
    values["COMMAND_SPACE"] = " "
    #
    # The character that indicates a comment.  While "(" is the most common,
    # ";" is also used.
    #
    values["COMMENT_SYMBOL"] = "("
    #
    # Default axis precision for metric is 3 digits after the decimal point.
    # (see http://linuxcnc.org/docs/2.7/html/gcode/overview.html#_g_code_best_practices)
    #
    values["DEFAULT_AXIS_PRECISION"] = 3
    #
    # The default precision for feed is also set to 3 for metric.
    #
    values["DEFAULT_FEED_PRECISION"] = 3
    #
    # Default axis precision for inch/imperial is 4 digits after the decimal point.
    #
    values["DEFAULT_INCH_AXIS_PRECISION"] = 4
    #
    # The default precision for feed is also set to 4 for inch/imperial.
    #
    values["DEFAULT_INCH_FEED_PRECISION"] = 4
    #
    # If TRANSLATE_DRILL_CYCLES is True, these are the drill cycles
    # that get translated to G0 and G1 commands.
    #
    values["DRILL_CYCLES_TO_TRANSLATE"] = ["G73", "G81", "G82", "G83"]
    #
    # If this is set to True, then M7, M8, and M9 commands
    # to enable/disable coolant will be output.
    #
    values["ENABLE_COOLANT"] = False
    #
    # If this is set to True, then commands that are placed in
    # comments that look like (MC_RUN_COMMAND: blah) will be output.
    #
    values["ENABLE_MACHINE_SPECIFIC_COMMANDS"] = False
    #
    # By default the line ending characters of the output file(s)
    # are written to match the system that the postprocessor runs on.
    # This is indicated by a value of "\n".  If you need to force the line ending
    # characters to a specific value, set this variable to "\n\n" (meaning "use \n"),
    # "\r" or "\r\n" instead.
    #
    values["END_OF_LINE_CHARACTERS"] = "\n"
    #
    # The starting precision for feed is also set to 3 digits after the decimal point.
    #
    values["FEED_PRECISION"] = 3
    #
    # This value shows up in the post_op comment as "Finish operation:".
    # At least one postprocessor changes it to "End" to produce "End operation:".
    #
    values["FINISH_LABEL"] = "Finish"
    #
    # The line number increment value
    #
    values["LINE_INCREMENT"] = 10
    #
    # The line number starting value
    #
    values["line_number"] = 100
    #
    # If this value is True, then a list of tool numbers
    # with their labels are output just before the preamble.
    #
    values["LIST_TOOLS_IN_PREAMBLE"] = False
    #
    # The name of the machine the postprocessor is for
    #
    values["MACHINE_NAME"] = "unknown machine"
    #
    # If this value is true G-code commands are suppressed if they are
    # the same as the previous line.
    #
    values["MODAL"] = False
    #
    # This defines the motion commands that might change the X, Y, and Z position.
    #
    values["MOTION_COMMANDS"] = [
        "G0",
        "G00",
        "G1",
        "G01",
        "G2",
        "G02",
        "G3",
        "G03",
    ]
    #
    # Keeps track of the motion mode currently in use.
    # G90 for absolute moves, G91 for relative
    #
    values["MOTION_MODE"] = "G90"
    #
    # If True enables special processing for operations with "Adaptive" in the name
    #
    values["OUTPUT_ADAPTIVE"] = False
    #
    # If True adds bCNC operation block headers to the output G-code file.
    #
    values["OUTPUT_BCNC"] = False
    #
    # If True output comments.  If False comments are suppressed.
    #
    values["OUTPUT_COMMENTS"] = True
    #
    # If True output blank lines.  If False blank lines are suppressed.
    #
    values["OUTPUT_BLANK_LINES"] = True
    #
    # if False duplicate axis values or feeds are suppressed
    # if they are the same as the previous line.
    #
    values["OUTPUT_DOUBLES"] = True
    #
    # If True output the machine name in the pre_op
    #
    values["OUTPUT_MACHINE_NAME"] = False
    #
    # If True output a header at the front of the G-code file.
    # The header contains comments similar to:
    #   (Exported by FreeCAD)
    #   (Post Processor: centroid_post)
    #   (Cam File: box.fcstd)
    #   (Output Time:2020-01-01 01:02:03.123456)
    #
    values["OUTPUT_HEADER"] = True
    #
    # If True output line numbers at the front of each line.
    # If False do not output line numbers.
    #
    values["OUTPUT_LINE_NUMBERS"] = False
    #
    # If True output Path labels at the beginning of each Path.
    #
    values["OUTPUT_PATH_LABELS"] = False
    #
    # If True output tool change G-code for M6 commands followed
    # by any commands in the "TOOL_CHANGE" value.
    # If False output the M6 command as a comment and do not output
    # any commands in the "TOOL_CHANGE" value.
    #
    values["OUTPUT_TOOL_CHANGE"] = True
    #
    # This dictionary/hash holds the functions that are used
    # to process the G-code parameter values
    #
    values["PARAMETER_FUNCTIONS"] = {}
    PostUtilsParse.init_parameter_functions(values["PARAMETER_FUNCTIONS"])
    #
    # This list controls the order of parameters in a line during output.
    #
    values["PARAMETER_ORDER"] = [
        "D",
        "H",
        "L",
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "U",
        "V",
        "W",
        "I",
        "J",
        "K",
        "R",
        "P",
        "E",
        "Q",
        "F",
        "S",
        "T",
    ]
    #
    # Any commands in this value will be output as the last commands
    # in the G-code file.
    #
    values["POSTAMBLE"] = """"""
    #
    # Any commands in this value will be output after the operation(s).
    #
    values["POST_OPERATION"] = """"""
    #
    # Any commands in this value will be output after the header and
    # safety block at the beginning of the G-code file.
    #
    values["PREAMBLE"] = """"""
    #
    # Any commands in this value will be output before the operation(s).
    #
    values["PRE_OPERATION"] = """"""
    #
    # Defines which G-code commands are considered "rapid" moves.
    #
    values["RAPID_MOVES"] = ["G0", "G00"]
    #
    # Any commands in this value are output after the operation(s)
    # and post_operation commands are output but before the
    # TOOLRETURN, SAFETYBLOCK, and POSTAMBLE.
    #
    values["RETURN_TO"] = None
    #
    # Any commands in this value are output after the header but before the preamble,
    # then again after the TOOLRETURN but before the POSTAMBLE.
    #
    values["SAFETYBLOCK"] = """"""
    #
    # If True then the G-code editor widget is shown before writing
    # the G-code to the file.
    #
    values["SHOW_EDITOR"] = True
    #
    # If True then the current machine units are output just before the PRE_OPERATION.
    #
    values["SHOW_MACHINE_UNITS"] = True
    #
    # If True then the current operation label is output just before the PRE_OPERATION
    # and just before the POST_OPERATION.
    #
    values["SHOW_OPERATION_LABELS"] = True
    #
    # The number of decimal places to use when outputting the speed (S) parameter.
    #
    values["SPINDLE_DECIMALS"] = 0
    #
    # The amount of time (in seconds) to wait after turning on the spindle
    # using an M3 or M4 command (a floating point number).
    #
    values["SPINDLE_WAIT"] = 0.0
    #
    # If true then then an M5 command to stop the spindle is output
    # after the M6 tool change command and before the TOOL_CHANGE commands.
    #
    values["STOP_SPINDLE_FOR_TOOL_CHANGE"] = True
    #
    # These commands are ignored by commenting them out.
    # Used when replacing the drill commands by G0 and G1 commands, for example.
    #
    values["SUPPRESS_COMMANDS"] = []
    #
    # Any commands in this value are output after the M6 command
    # when changing at tool (if OUTPUT_TOOL_CHANGE is True).
    #
    values["TOOL_CHANGE"] = """"""
    #
    # Any commands in this value are output after the POST_OPERATION,
    # RETURN_TO, and OUTPUT_BCNC and before the SAFETYBLOCK and POSTAMBLE.
    #
    values["TOOLRETURN"] = """"""
    #
    # If true, G81, G82 & G83 drill moves are translated into G0/G1 moves.
    #
    values["TRANSLATE_DRILL_CYCLES"] = False
    #
    # These values keep track of whether we are in Metric mode (G21)
    # or inches/imperial mode (G20).
    #
    values["UNITS"] = "G21"
    values["UNIT_FORMAT"] = "mm"
    values["UNIT_SPEED_FORMAT"] = "mm/min"
    #
    # If true a tool length command (G43) will be output following tool changes.
    #
    values["USE_TLO"] = True


def process_shared_arguments(
    values: Values,
    parser: Parser,
    argstring: str,
    all_visible: Parser,
    filename: str,
) -> Tuple[bool, Union[None, str, argparse.Namespace]]:
    """Process the arguments to the postprocessor."""
    args: argparse.Namespace
    argument_text: str
    v: str

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.output_all_arguments:
            argument_text = all_visible.format_help()
            if not filename == "-":
                with open(
                    filename,
                    "w",
                    encoding="utf-8",
                ) as f:
                    f.write(argument_text)
            return (False, argument_text)
        if args.output_visible_arguments:
            argument_text = parser.format_help()
            if not filename == "-":
                with open(
                    filename,
                    "w",
                    encoding="utf-8",
                ) as f:
                    f.write(argument_text)
            return (False, argument_text)
        # Default to metric unless an argument overrides it
        values["UNITS"] = "G21"
        if args.metric:
            values["UNITS"] = "G21"
        if args.inches:
            values["UNITS"] = "G20"
        if values["UNITS"] == "G21":
            values["UNIT_FORMAT"] = "mm"
            values["UNIT_SPEED_FORMAT"] = "mm/min"
        elif values["UNITS"] == "G20":
            values["UNIT_FORMAT"] = "in"
            values["UNIT_SPEED_FORMAT"] = "in/min"
        # The precision-related arguments need to be processed
        # after the metric/inches arguments are processed.
        # If both --axis-precision and --precision are given,
        # the --axis-precision argument "wins".
        if args.axis_precision != -1:
            values["AXIS_PRECISION"] = args.axis_precision
        elif args.precision != -1:
            values["AXIS_PRECISION"] = args.precision
        else:
            if values["UNITS"] == "G21":
                values["AXIS_PRECISION"] = values["DEFAULT_AXIS_PRECISION"]
            if values["UNITS"] == "G20":
                values["AXIS_PRECISION"] = values["DEFAULT_INCH_AXIS_PRECISION"]
        # If both --feed-precision and --precision are given,
        # the --feed-precision argument "wins".
        if args.feed_precision != -1:
            values["FEED_PRECISION"] = args.feed_precision
        elif args.precision != -1:
            values["FEED_PRECISION"] = args.precision
        else:
            if values["UNITS"] == "G21":
                values["FEED_PRECISION"] = values["DEFAULT_FEED_PRECISION"]
            elif values["UNITS"] == "G20":
                values["FEED_PRECISION"] = values["DEFAULT_INCH_FEED_PRECISION"]
        if args.axis_modal:
            values["OUTPUT_DOUBLES"] = False
        if args.no_axis_modal:
            values["OUTPUT_DOUBLES"] = True
        if args.bcnc:
            values["OUTPUT_BCNC"] = True
        if args.no_bcnc:
            values["OUTPUT_BCNC"] = False
        if args.chipbreaking_amount:
            values["CHIPBREAKING_AMOUNT"] = Units.parseQuantity(args.chipbreaking_amount)
        values["COMMAND_SPACE"] = args.command_space
        if args.comments:
            values["OUTPUT_COMMENTS"] = True
        if args.no_comments:
            values["OUTPUT_COMMENTS"] = False
        if args.comment_symbol:
            values["COMMENT_SYMBOL"] = args.comment_symbol
        if args.enable_coolant:
            values["ENABLE_COOLANT"] = True
        if args.disable_coolant:
            values["ENABLE_COOLANT"] = False
        if args.enable_machine_specific_commands:
            values["ENABLE_MACHINE_SPECIFIC_COMMANDS"] = True
        if args.disable_machine_specific_commands:
            values["ENABLE_MACHINE_SPECIFIC_COMMANDS"] = False
        if args.end_of_line_characters:
            if args.end_of_line_characters == "\\n" or args.end_of_line_characters == "\n":
                values["END_OF_LINE_CHARACTERS"] = "\n\n"
            elif args.end_of_line_characters == "\\r" or args.end_of_line_characters == "\r":
                values["END_OF_LINE_CHARACTERS"] = "\r"
            elif args.end_of_line_characters == "\\r\\n" or args.end_of_line_characters == "\r\n":
                values["END_OF_LINE_CHARACTERS"] = "\r\n"
            else:
                print("invalid end_of_line_characters, ignoring")
        if args.finish_label:
            values["FINISH_LABEL"] = args.finish_label
        if args.header:
            values["OUTPUT_HEADER"] = True
        if args.no_header:
            values["OUTPUT_HEADER"] = False
        if args.line_number_increment != -1:
            values["LINE_INCREMENT"] = args.line_number_increment
        if args.line_number_start != -1:
            values["line_number"] = args.line_number_start
        if args.line_numbers:
            values["OUTPUT_LINE_NUMBERS"] = True
        if args.no_line_numbers:
            values["OUTPUT_LINE_NUMBERS"] = False
        if args.list_tools_in_preamble:
            values["LIST_TOOLS_IN_PREAMBLE"] = True
        if args.no_list_tools_in_preamble:
            values["LIST_TOOLS_IN_PREAMBLE"] = False
        if args.modal:
            values["MODAL"] = True
        if args.no_modal:
            values["MODAL"] = False
        if args.output_machine_name:
            values["OUTPUT_MACHINE_NAME"] = True
        if args.no_output_machine_name:
            values["OUTPUT_MACHINE_NAME"] = False
        if args.output_path_labels:
            values["OUTPUT_PATH_LABELS"] = True
        if args.no_output_path_labels:
            values["OUTPUT_PATH_LABELS"] = False
        if args.postamble is not None:
            values["POSTAMBLE"] = args.postamble.replace("\\n", "\n")
        if args.post_operation is not None:
            values["POST_OPERATION"] = args.post_operation.replace("\\n", "\n")
        if args.preamble is not None:
            values["PREAMBLE"] = args.preamble.replace("\\n", "\n")
        if args.return_to != "":
            values["RETURN_TO"] = [int(v) for v in args.return_to.split(",")]
            if len(values["RETURN_TO"]) != 3:
                values["RETURN_TO"] = None
                print("--return-to coordinates must be specified as <x>,<y>,<z>, ignoring")
        if args.show_editor:
            values["SHOW_EDITOR"] = True
        if args.no_show_editor:
            values["SHOW_EDITOR"] = False
        if args.tlo:
            values["USE_TLO"] = True
        if args.no_tlo:
            values["USE_TLO"] = False
        if args.tool_change:
            values["OUTPUT_TOOL_CHANGE"] = True
        if args.no_tool_change:
            values["OUTPUT_TOOL_CHANGE"] = False
        if args.translate_drill:
            values["TRANSLATE_DRILL_CYCLES"] = True
        if args.no_translate_drill:
            values["TRANSLATE_DRILL_CYCLES"] = False
        if args.wait_for_spindle > 0.0:
            values["SPINDLE_WAIT"] = args.wait_for_spindle

    except (ArithmeticError, LookupError):
        return (False, None)

    return (True, args)


#
# LinuxCNC (and GRBL) G-Code Parameter/word Patterns
# __________________________________________________
#
# LinuxCNC words (called parameters in this code in many places) may be
# reordered in any way without changing the meaning of the line.
# However, the documentation shows the examples with the parameters/words
# in a certain order that people might be used to and want to see.
#
#   axes    one or more of "X Y Z A B C U V W", usually in that order
#
# default parameter order     D H L axes I J K R P E Q F S T $
#
# G10                         L P axes R I J Q
#
# G33.1                       X Y Z K I $
#
# G53 G00|G01 or G00|G01 G53  X Y Z
#
# G73, G74, G81 to G86, G89   "X Y Z" | "U V W" R Q L P F K $
#
# G76                         P Z I J R K Q H E L $
#
# M19                         R Q P $
#
# M66                         P|E L Q
#
# M98                         P Q L
#
# N and O are associated with line numbers
#
#
#
# Tormach PathPilot (based on LinuxCNC, mostly) G-Code Patterns
# _____________________________________________________________
#
# (Just the exceptions to the LinuxCNC patterns shown above)
#
# G47                         Z R X Y P Q D
#
#
# Mach4 G-Code Patterns
# _____________________
#
# (Just the exceptions to the LinuxCNC patterns shown above)
#
# G10                        L1 P Z W D R X U Y V Q
#
# G30                        P axes
#
# G41, G42                   D|P X Y F
#
# G65, G66                   P A B C
#
# G73, G74, G76, G81-9       X Y Z Q R I J P L F
#
# G84.2, G84.3               X Y Z R P L F J
#
#
#
# Centroid G-Code Patterns
# ________________________
#
# (Just the exceptions to the LinuxCNC patterns shown above)
#
# E1-E6 are equivalent to G54-G59
#
# G10                       P D H R
#
# G65                       P L args (args are A-Z excluding G, L, N, O, and P)
# G65                       "program.cnc" L args
#
# G117, G118, G119          P X Y Z I J K P Q
#



# ***************************************************************************
# *   Copyright (c) 2017 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM import tools"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

## @package importToolsFem
#  \ingroup FEM
#  \brief FreeCAD FEM import tools

import FreeCAD
from FreeCAD import Console


def get_FemMeshObjectMeshGroups(fem_mesh_obj):
    """
    Get mesh groups from mesh.
    """
    # this method is not really needed. It is used in Fenics mesh only.
    # there was an exception handling if there was no Group property, but
    # any FemMesh should have the Group property
    # if not it would be a bug SMESH

    return fem_mesh_obj.FemMesh.Groups


def get_FemMeshObjectOrder(fem_mesh_obj):
    """
    Gets element order. Element order counting based on number of nodes on
    edges. Edge with 2 nodes -> linear elements, Edge with 3 nodes ->
    quadratic elements, and so on. No edges in mesh -> not determined.
    (Is this possible? Seems to be a very degenerate case.)
    If there are edges with different number of nodes appearing, return
    list of orders.
    """
    presumable_order = None

    edges = fem_mesh_obj.FemMesh.Edges

    if edges != ():
        edges_length_set = list({len(fem_mesh_obj.FemMesh.getElementNodes(e)) for e in edges})
        # only need set to eliminate double entries

        if len(edges_length_set) == 1:
            presumable_order = edges_length_set[0] - 1
        else:
            presumable_order = [el - 1 for el in edges_length_set]
    else:
        Console.PrintMessage(
            "Found no edges in mesh: Element order determination does not work without them.\n"
        )

    return presumable_order


def get_FemMeshObjectDimension(fem_mesh_obj):
    """Count all entities in an abstract sense, to distinguish which dimension the mesh is
    (i.e. linemesh, facemesh, volumemesh)
    """
    dim = None

    if fem_mesh_obj.FemMesh.Nodes != ():
        dim = 0
    if fem_mesh_obj.FemMesh.Edges != ():
        dim = 1
    if fem_mesh_obj.FemMesh.Faces != ():
        dim = 2
    if fem_mesh_obj.FemMesh.Volumes != ():
        dim = 3

    return dim


def get_FemMeshObjectElementTypes(fem_mesh_obj, remove_zero_element_entries=True):
    """
    Spit out all elements in the mesh with their appropriate dimension.
    """
    FreeCAD_element_names_dims = {
        "Node": 0,
        "Edge": 1,
        "Hexa": 3,
        "Polygon": 2,
        "Polyhedron": 3,
        "Prism": 3,
        "Pyramid": 3,
        "Quadrangle": 2,
        "Tetra": 3,
        "Triangle": 2,
    }

    eval_dict = locals()  # to access local variables from eval
    elements_list_with_zero = [
        (eval("fem_mesh_obj.FemMesh." + s + "Count", eval_dict), s, d)
        for (s, d) in FreeCAD_element_names_dims.items()
    ]
    # ugly but necessary
    if remove_zero_element_entries:
        elements_list = [(num, s, d) for (num, s, d) in elements_list_with_zero if num > 0]
    else:
        elements_list = elements_list_with_zero

    return elements_list


def get_MaxDimElementFromList(elem_list):
    """
    Gets element with the maximal dimension in the mesh to determine cells.
    """
    elem_list.sort(key=lambda t: t[2])
    return elem_list[-1]


def make_femmesh(mesh_data):
    """makes an FreeCAD FEM Mesh object from FEM Mesh data"""
    import Fem

    mesh = Fem.FemMesh()
    m = mesh_data
    if ("Nodes" in m) and (len(m["Nodes"]) > 0):
        FreeCAD.Console.PrintLog("Found: nodes\n")
        if (
            ("Seg2Elem" in m)
            or ("Seg3Elem" in m)
            or ("Tria3Elem" in m)
            or ("Tria6Elem" in m)
            or ("Quad4Elem" in m)
            or ("Quad8Elem" in m)
            or ("Tetra4Elem" in m)
            or ("Tetra10Elem" in m)
            or ("Penta6Elem" in m)
            or ("Penta15Elem" in m)
            or ("Hexa8Elem" in m)
            or ("Hexa20Elem" in m)
        ):

            nds = m["Nodes"]
            FreeCAD.Console.PrintLog("Found: elements\n")
            for i in nds:
                n = nds[i]
                mesh.addNode(n[0], n[1], n[2], i)
            elms_hexa8 = m["Hexa8Elem"]
            for i in elms_hexa8:
                e = elms_hexa8[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7]], i)
            elms_penta6 = m["Penta6Elem"]
            for i in elms_penta6:
                e = elms_penta6[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5]], i)
            elms_tetra4 = m["Tetra4Elem"]
            for i in elms_tetra4:
                e = elms_tetra4[i]
                mesh.addVolume([e[0], e[1], e[2], e[3]], i)
            elms_tetra10 = m["Tetra10Elem"]
            for i in elms_tetra10:
                e = elms_tetra10[i]
                mesh.addVolume([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7], e[8], e[9]], i)
            elms_penta15 = m["Penta15Elem"]
            for i in elms_penta15:
                e = elms_penta15[i]
                mesh.addVolume(
                    [
                        e[0],
                        e[1],
                        e[2],
                        e[3],
                        e[4],
                        e[5],
                        e[6],
                        e[7],
                        e[8],
                        e[9],
                        e[10],
                        e[11],
                        e[12],
                        e[13],
                        e[14],
                    ],
                    i,
                )
            elms_hexa20 = m["Hexa20Elem"]
            for i in elms_hexa20:
                e = elms_hexa20[i]
                mesh.addVolume(
                    [
                        e[0],
                        e[1],
                        e[2],
                        e[3],
                        e[4],
                        e[5],
                        e[6],
                        e[7],
                        e[8],
                        e[9],
                        e[10],
                        e[11],
                        e[12],
                        e[13],
                        e[14],
                        e[15],
                        e[16],
                        e[17],
                        e[18],
                        e[19],
                    ],
                    i,
                )
            elms_tria3 = m["Tria3Elem"]
            for i in elms_tria3:
                e = elms_tria3[i]
                mesh.addFace([e[0], e[1], e[2]], i)
            elms_tria6 = m["Tria6Elem"]
            for i in elms_tria6:
                e = elms_tria6[i]
                mesh.addFace([e[0], e[1], e[2], e[3], e[4], e[5]], i)
            elms_quad4 = m["Quad4Elem"]
            for i in elms_quad4:
                e = elms_quad4[i]
                mesh.addFace([e[0], e[1], e[2], e[3]], i)
            elms_quad8 = m["Quad8Elem"]
            for i in elms_quad8:
                e = elms_quad8[i]
                mesh.addFace([e[0], e[1], e[2], e[3], e[4], e[5], e[6], e[7]], i)
            elms_seg2 = m["Seg2Elem"]
            for i in elms_seg2:
                e = elms_seg2[i]
                mesh.addEdge([e[0], e[1]], i)
            elms_seg3 = m["Seg3Elem"]
            for i in elms_seg3:
                e = elms_seg3[i]
                mesh.addEdge([e[0], e[1], e[2]], i)
            Console.PrintLog(
                "imported mesh: {} nodes, {} HEXA8, {} PENTA6, {} TETRA4, {} TETRA10, {} PENTA15\n".format(
                    len(nds),
                    len(elms_hexa8),
                    len(elms_penta6),
                    len(elms_tetra4),
                    len(elms_tetra10),
                    len(elms_penta15),
                )
            )
            Console.PrintLog(
                "imported mesh: {} "
                "HEXA20, {} TRIA3, {} TRIA6, {} QUAD4, {} QUAD8, {} SEG2, {} SEG3\n".format(
                    len(elms_hexa20),
                    len(elms_tria3),
                    len(elms_tria6),
                    len(elms_quad4),
                    len(elms_quad8),
                    len(elms_seg2),
                    len(elms_seg3),
                )
            )
        else:
            Console.PrintError("No Elements found!\n")
    else:
        Console.PrintError("No Nodes found!\n")
    return mesh


def make_dict_from_femmesh(femmesh):
    """
    Converts FemMesh into dictionary structure which can immediately used
    from importToolsFem.make_femmesh(mesh_data) to create a valid FEM mesh.
    """
    # this dict can be easily saved and reloaded by yaml
    # see importYamlJasonMesh for a implementation

    mesh_data = {}

    seg2 = []
    seg3 = []

    tri3 = []
    tri6 = []
    quad4 = []
    quad8 = []

    tet4 = []
    tet10 = []
    hex8 = []
    hex20 = []
    pent6 = []
    pent15 = []

    # associations for lengths of tuples to different
    # edge, face, and volume elements

    len_to_edge = {2: seg2, 3: seg3}
    len_to_face = {3: tri3, 6: tri6, 4: quad4, 8: quad8}
    len_to_volume = {4: tet4, 10: tet10, 8: hex8, 20: hex20, 6: pent6, 15: pent15}

    # analyze edges

    for e in femmesh.Edges:
        t = femmesh.getElementNodes(e)
        len_to_edge[len(t)].append((e, t))

    # analyze faces

    for f in femmesh.Faces:
        t = femmesh.getElementNodes(f)
        len_to_face[len(t)].append((f, t))

    # analyze volumes

    for v in femmesh.Volumes:
        t = femmesh.getElementNodes(v)
        len_to_volume[len(t)].append((v, t))

    mesh_data = {
        "Nodes": {k: (v.x, v.y, v.z) for (k, v) in femmesh.Nodes.items()},
        "Seg2Elem": dict(seg2),
        "Seg3Elem": dict(seg3),
        "Tria3Elem": dict(tri3),
        "Tria6Elem": dict(tri6),
        "Quad4Elem": dict(quad4),
        "Quad8Elem": dict(quad8),
        "Tetra4Elem": dict(tet4),
        "Tetra10Elem": dict(tet10),
        "Hexa8Elem": dict(hex8),
        "Hexa20Elem": dict(hex20),
        "Penta6Elem": dict(pent6),
        "Penta15Elem": dict(pent15),
        "Groups": {
            group_num: (femmesh.getGroupName(group_num), femmesh.getGroupElements(group_num))
            for group_num in femmesh.Groups
        },
    }
    # no pyr5, pyr13?
    # no groups?
    return mesh_data


def fill_femresult_mechanical(res_obj, result_set):
    """fills a FreeCAD FEM mechanical result object with result data"""
    if "number" in result_set:
        eigenmode_number = result_set["number"]
    else:
        eigenmode_number = 0

    if "time" in result_set:
        step_time = result_set["time"]
        step_time = round(step_time, 2)

    disp = None
    # if disp exists, fill res_obj.NodeNumbers and
    # res_obj.DisplacementVectors as well as stress and strain
    # furthermore the eigenmode number
    if "disp" in result_set:
        disp = result_set["disp"]
        res_obj.DisplacementVectors = list(map((lambda x: x), disp.values()))
        res_obj.NodeNumbers = list(disp)

        # fill res_obj.NodeStressXX etc if they exist in result_set
        # list values are just added
        # Should we check if the key in stress and strain dict
        # is the same as the number in NodeNumbers?
        if "stress" in result_set:
            stress = result_set["stress"]
            Sxx = []
            Syy = []
            Szz = []
            Sxy = []
            Sxz = []
            Syz = []
            # values_S .. stress_tensor .. (Sxx, Syy, Szz, Sxy, Sxz, Syz)
            for i, values_S in enumerate(stress.values()):
                Sxx.append(values_S[0])
                Syy.append(values_S[1])
                Szz.append(values_S[2])
                Sxy.append(values_S[3])
                Sxz.append(values_S[4])
                Syz.append(values_S[5])
            res_obj.NodeStressXX = Sxx
            res_obj.NodeStressYY = Syy
            res_obj.NodeStressZZ = Szz
            res_obj.NodeStressXY = Sxy
            res_obj.NodeStressXZ = Sxz
            res_obj.NodeStressYZ = Syz

        # fill res_obj.NodeStrainXX etc if they exist in result_set
        if "strain" in result_set:
            strain = result_set["strain"]
            Exx = []
            Eyy = []
            Ezz = []
            Exy = []
            Exz = []
            Eyz = []
            # values_E .. straintuple .. (Exx, Eyy, Ezz, Exy, Exz, Eyz)
            for i, values_E in enumerate(strain.values()):
                Exx.append(values_E[0])
                Eyy.append(values_E[1])
                Ezz.append(values_E[2])
                Exy.append(values_E[3])
                Exz.append(values_E[4])
                Eyz.append(values_E[5])
            res_obj.NodeStrainXX = Exx
            res_obj.NodeStrainYY = Eyy
            res_obj.NodeStrainZZ = Ezz
            res_obj.NodeStrainXY = Exy
            res_obj.NodeStrainXZ = Exz
            res_obj.NodeStrainYZ = Eyz

        # fill Equivalent Plastic strain if they exist
        if "peeq" in result_set:
            Peeq = result_set["peeq"]
            if len(Peeq) > 0:
                if len(Peeq.values()) != len(disp.values()):
                    # how is this possible? An example is needed!
                    Console.PrintError("PEEQ seems to have extra nodes.\n")
                    Pe = []
                    Pe_extra_nodes = list(Peeq.values())
                    nodes = len(disp.values())
                    for i in range(nodes):
                        Pe_value = Pe_extra_nodes[i]
                        Pe.append(Pe_value)
                    res_obj.Peeq = Pe
                else:
                    res_obj.Peeq = list(Peeq.values())

        # fill eigenmode number if they exist
        if eigenmode_number > 0:
            res_obj.Eigenmode = eigenmode_number

    # it is assumed Temperature can not exist without disp
    # TODO really proof this
    # if temperature can exist without disp:
    # move them out of disp if conditiona and set NodeNumbers
    if "temp" in result_set:
        Temperature = result_set["temp"]
        if len(Temperature) > 0:
            if disp is None:
                res_obj.Temperature = list(Temperature.values())
                res_obj.NodeNumbers = list(Temperature.keys())
            elif len(Temperature.values()) != len(disp.values()):
                Temp = []
                Temp_extra_nodes = list(Temperature.values())
                nodes = len(disp.values())
                for i in range(nodes):
                    # how is this possible? An example is needed!
                    Console.PrintError("Temperature seems to have extra nodes.\n")
                    Temp_value = Temp_extra_nodes[i]
                    Temp.append(Temp_value)
                res_obj.Temperature = list(map((lambda x: x), Temp))
            else:
                res_obj.Temperature = list(map((lambda x: x), Temperature.values()))
            res_obj.Time = step_time

    if "heatflux" in result_set:
        HeatFlux = result_set["heatflux"]
        if HeatFlux:
            res_obj.HeatFlux = list(map((lambda x: x), HeatFlux.values()))

    # fill res_obj.MassFlow
    if "mflow" in result_set:
        MassFlow = result_set["mflow"]
        if len(MassFlow) > 0:
            res_obj.MassFlowRate = list(map((lambda x: x), MassFlow.values()))
            res_obj.Time = step_time
            # disp does not exist, res_obj.NodeNumbers needs to be set
            res_obj.NodeNumbers = list(MassFlow)

    # fill res_obj.NetworkPressure, disp does not exist, see MassFlow
    if "npressure" in result_set:
        NetworkPressure = result_set["npressure"]
        if len(NetworkPressure) > 0:
            res_obj.NetworkPressure = list(map((lambda x: x), NetworkPressure.values()))
            res_obj.Time = step_time

    return res_obj



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from Path.Tool.toolbit import ToolBit
import Path.Tool.Controller as PathToolController
from CAMTests.PathTestUtils import PathTestBase


class TestPathToolController(PathTestBase):
    def setUp(self):
        self.doc = FreeCAD.newDocument("TestPathToolController")
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "True")

    def tearDown(self):
        FreeCAD.closeDocument(self.doc.Name)
        FreeCAD.ConfigSet("SuppressRecomputeRequiredDialog", "")

    def createTool(self, name="t1", diameter=1.75):
        attrs = {
            "name": name or "t1",
            "shape": "endmill.fcstd",
            "parameter": {"Diameter": diameter},
            "attribute": {},
        }
        print(f"Debug: attrs['attribute'] is {attrs['attribute']}")
        toolbit = ToolBit.from_dict(attrs)
        return toolbit.attach_to_doc(doc=FreeCAD.ActiveDocument)

    def test00(self):
        """Verify ToolController templateAttrs"""
        t = self.createTool("T1")
        tc = PathToolController.Create("TC0", t)

        tc.Label = "ToolController"
        tc.ToolNumber = 7
        tc.VertFeed = "3 in/s"
        tc.VertFeed = round(tc.VertFeed, 1)
        tc.HorizFeed = "10 mm/s"
        tc.VertRapid = 40
        tc.HorizRapid = 28
        tc.SpindleDir = "Reverse"
        tc.SpindleSpeed = 12000

        attrs = tc.Proxy.templateAttrs(tc)

        self.assertEqual(attrs["name"], "TC0")
        self.assertEqual(attrs["label"], "ToolController")
        self.assertEqual(attrs["nr"], 7)
        self.assertEqual(attrs["vfeed"], "76.2 mm/s")
        self.assertEqual(attrs["hfeed"], "10.0 mm/s")
        self.assertEqual(attrs["vrapid"], "40.0 mm/s")
        self.assertEqual(attrs["hrapid"], "28.0 mm/s")
        self.assertEqual(attrs["dir"], "Reverse")
        self.assertEqual(attrs["speed"], 12000)
        self.assertEqual(attrs["tool"], t.Proxy.to_dict())

        return tc

    def test01(self):
        """Verify ToolController template roundtrip."""

        tc0 = self.test00()
        tc1 = PathToolController.FromTemplate(tc0.Proxy.templateAttrs(tc0))

        self.assertNotEqual(tc0.Name, tc1.Name)
        self.assertNotEqual(tc0.Label, tc1.Label)
        self.assertEqual(tc0.ToolNumber, tc1.ToolNumber)
        self.assertRoughly(tc0.VertFeed, tc1.VertFeed)
        self.assertRoughly(tc0.HorizFeed, tc1.HorizFeed)
        self.assertRoughly(tc0.VertRapid, tc1.VertRapid)
        self.assertRoughly(tc0.HorizRapid, tc1.HorizRapid)
        self.assertEqual(tc0.SpindleDir, tc1.SpindleDir)
        self.assertRoughly(tc0.SpindleSpeed, tc1.SpindleSpeed)
        # These are not valid because the name & label get adjusted if there
        # is a conflict. No idea how this could work with the C implementation
        # self.assertEqual(tc0.Tool.Name, tc1.Tool.Name)
        # self.assertEqual(tc0.Tool.Label, tc1.Tool.Label)
        self.assertRoughly(tc0.Tool.Diameter, tc1.Tool.Diameter)



# SPDX-License-Identifier: LGPL-2.1-or-later

"""

CAM post processor for CNC machines with a Fanuc controller.

"""

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2021 shadowbane1000 <tyler@colberts.us>                 *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
import os.path
import Path.Base.Util as PathUtil
import Path.Post.Utils as PostUtils
from PathScripts import PathUtils
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable should be suitable for most Fanuc controllers.
It has only been tested on a 21i-MB controller on a 3 axis mill.
This postprocessor, once placed in the appropriate PathScripts folder,
can be used directly from inside FreeCAD, via the GUI importer or via
python scripts with:

import fanuc_post
fanuc_post.export(object,"/path/to/file.ncc","")
"""

# Preamble text will appear at the beginning of the GCODE output file.
DEFAULT_PREAMBLE = """G17 G54 G40 G49 G80 G90 G94
"""

# Postamble text will appear following the last operation.
DEFAULT_POSTAMBLE = """M05
G17 G54 G90 G80 G40
M30
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="fanuc", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--precision", help="number of digits of precision, default=3 (mm) or 4 (in)")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="'
    + DEFAULT_PREAMBLE.replace("\n", "\\n")
    + '"',
)
parser.add_argument(
    "--postamble",
    help="set commands to be issued after the last command, "
    + 'default="'
    + DEFAULT_POSTAMBLE.replace("\n", "\\n")
    + '"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)
parser.add_argument(
    "--no-modal",
    action="store_true",
    help="Don't output the Same G-command Name USE NonModal Mode",
)
parser.add_argument(
    "--no-axis-modal", action="store_true", help="Don't output the Same Axis Value Mode"
)
parser.add_argument(
    "--no-tlo",
    action="store_true",
    help="suppress tool length offset (G43) following tool changes",
)
parser.add_argument(
    "--end-spindle-empty",
    action="store_true",
    help="place last tool in tool change carousel before postamble",
)

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = True  # if true commands are suppressed if the same as previous line.
USE_TLO = True  # if true G43 will be output following tool changes
OUTPUT_DOUBLES = (
    False  # if false duplicate axis values are suppressed if the same as previous line.
)
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

END_SPINDLE_EMPTY = False

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "fanuc"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}
PRECISION = 3

# this global is used to pass spindle speed from the tool command into the machining command for
# rigid tapping.
tapSpeed = 0

PREAMBLE = DEFAULT_PREAMBLE
POSTAMBLE = DEFAULT_POSTAMBLE

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
# Move to tool change Z position
TOOL_CHANGE = """G28 G91 Z0
"""

# List of drill G codes where some parameters are required and their
# required parameters.
DRILL_OPERATION = ("G73", "G81", "G82", "G83", "G85")
DRILL_PARAM_REQ = ("L", "P", "Q", "R", "Z")

# The settings shared between methods
PREAMBLE = None
POSTAMBLE = None


def processArguments(argstring):
    """
    Apply default values and command line arguments before
    processing commands.

    """
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global DEFAULT_PREAMBLE
    global DEFAULT_POSTAMBLE
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global END_SPINDLE_EMPTY
    global OUTPUT_DOUBLES
    global LINENR

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        else:
            OUTPUT_HEADER = True
        if args.no_comments:
            OUTPUT_COMMENTS = False
        else:
            OUTPUT_COMMENTS = True
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
            LINENR = 100
        else:
            OUTPUT_LINE_NUMBERS = False
        if args.no_show_editor:
            SHOW_EDITOR = False
        else:
            SHOW_EDITOR = True
        # print("Show editor = %s" % SHOW_EDITOR)  # Commented to reduce test noise
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        else:
            PREAMBLE = DEFAULT_PREAMBLE
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        else:
            POSTAMBLE = DEFAULT_POSTAMBLE
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4
        else:
            UNITS = "G21"
            UNIT_SPEED_FORMAT = "mm/min"
            UNIT_FORMAT = "mm"
            PRECISION = 3
        if args.precision:
            PRECISION = int(args.precision)
        if args.no_modal:
            MODAL = False
        else:
            MODAL = True
        if args.no_tlo:
            USE_TLO = False
        else:
            USE_TLO = True
        if args.no_axis_modal:
            OUTPUT_DOUBLES = True
        else:
            OUTPUT_DOUBLES = False
        if args.end_spindle_empty:
            END_SPINDLE_EMPTY = True
        else:
            END_SPINDLE_EMPTY = False

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT
    global HORIZRAPID
    global VERTRAPID

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return None

    # print("postprocessing...")  # Commented to reduce test noise
    gcode = ""

    gcode += "%\n"

    # write header
    if OUTPUT_HEADER:
        # Get current version info
        major = int(FreeCAD.ConfigGet("BuildVersionMajor"))
        minor = int(FreeCAD.ConfigGet("BuildVersionMinor"))

        # the filename variable always contain "-", use more relevant
        # information
        job = PathUtils.findParentJob(objectslist[0])
        if job:
            body, job = job.FullName.split("#")
        else:
            # Workaround for the TestFanucPost code, where there is no
            # job returned by findParentJob
            body, job = ("FREECAD-FILENAME-GOES-HERE", "JOB-NAME-GOES-HERE")
        gcode += "(" + body.upper() + ", " + job.upper() + ")\n"
        gcode += (
            linenumber() + "(POST PROCESSOR: FANUC USING FREECAD %d.%d" % (major, minor) + ")\n"
        )
        gcode += linenumber() + "(OUTPUT TIME:" + str(now).upper() + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(BEGIN PREAMBLE)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:

        # to stay compatible with FreeCAD 1.0
        def activeForOp(obj):
            # The activeForOp method is available since 2025-05-04 /
            # commit 1e87d8e6681b755b9757f94b1201e50eb84b28a2
            if hasattr(PathUtil, "activeForOp"):
                return PathUtil.activeForOp(obj)
            if hasattr(obj, "Active"):
                return obj.Active
            if hasattr(obj, "Base") and hasattr(obj.Base, "Active"):
                return obj.Base.Active
            return True

        # Skip inactive operations
        if not activeForOp(obj):
            continue

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(BEGIN OPERATION: %s)\n" % obj.Label.upper()
            gcode += linenumber() + "(MACHINE UNITS: %s)\n" % (UNIT_SPEED_FORMAT.upper())
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # to stay compatible with FreeCAD 1.0
        def coolantModeForOp(obj):
            # The coolantModeForOp method is available since
            # 2025-05-04 / commit
            # 1e87d8e6681b755b9757f94b1201e50eb84b28a2
            if hasattr(PathUtil, "coolantModeForOp"):
                return PathUtil.coolantModeForOp(obj)
            if (
                hasattr(obj, "CoolantMode")
                or hasattr(obj, "Base")
                and hasattr(obj.Base, "CoolantMode")
            ):
                if hasattr(obj, "CoolantMode"):
                    return obj.CoolantMode
                else:
                    return obj.Base.CoolantMode
            return "None"

        # get coolant mode
        coolantMode = coolantModeForOp(obj)

        # turn coolant on if required
        if OUTPUT_COMMENTS:
            if not coolantMode == "None":
                gcode += linenumber() + "(COOLANT ON:" + coolantMode.upper() + ")\n"
        if coolantMode == "Flood":
            gcode += linenumber() + "M8" + "\n"
        if coolantMode == "Mist":
            gcode += linenumber() + "M7" + "\n"

        # process the operation gcode
        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(FINISH OPERATION: %s)\n" % obj.Label.upper()
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # turn coolant off if required
        if not coolantMode == "None":
            if OUTPUT_COMMENTS:
                gcode += linenumber() + "(COOLANT OFF:" + coolantMode.upper() + ")\n"
            gcode += linenumber() + "M9" + "\n"

    if END_SPINDLE_EMPTY:
        if OUTPUT_COMMENTS:
            gcode += "(BEGIN MAKING SPINDLE EMPTY)\n"
        gcode += linenumber() + "M05\n"
        for line in TOOL_CHANGE.splitlines(True):
            gcode += linenumber() + line
        gcode += linenumber() + "M6 T0\n"
    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "(BEGIN POSTAMBLE)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if FreeCAD.GuiUp and SHOW_EDITOR:
        dia = PostUtils.GCodeEditorDialog()
        dia.editor.setText(gcode)
        result = dia.exec_()
        if result:
            final = dia.editor.toPlainText()
        else:
            final = gcode
    else:
        final = gcode

    # print("done postprocessing.")  # Commented to reduce test noise

    if not filename == "-":
        gfile = pyopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    global PRECISION
    global DRILL_OPERATION
    global DRILL_PARAM_REQ
    global MODAL
    global OUTPUT_DOUBLES
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT
    global tapSpeed

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"
    currLocation = {}  # keep track for no doubles
    print("Startup!")

    # the order of parameters
    # arcs need work.  original code from mach3_4 doesn't want K properties on XY plane.  Not sure
    # what fanuc does here.
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    firstmove = Path.Command("G0", {"X": -1, "Y": -1, "Z": -1, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path

        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(" + pathobj.Label + ")\n"

        adaptiveOp = False
        opHorizRapid = 0
        opVertRapid = 0

        if "Adaptive" in pathobj.Name:
            adaptiveOp = True
            if hasattr(pathobj, "ToolController"):
                if (
                    hasattr(pathobj.ToolController, "HorizRapid")
                    and pathobj.ToolController.HorizRapid > 0
                ):
                    opHorizRapid = Units.Quantity(
                        pathobj.ToolController.HorizRapid, FreeCAD.Units.Velocity
                    )
                else:
                    FreeCAD.Console.PrintWarning(
                        "Tool Controller Horizontal Rapid Values are unset" + "\n"
                    )

                if (
                    hasattr(pathobj.ToolController, "VertRapid")
                    and pathobj.ToolController.VertRapid > 0
                ):
                    opVertRapid = Units.Quantity(
                        pathobj.ToolController.VertRapid, FreeCAD.Units.Velocity
                    )
                else:
                    FreeCAD.Console.PrintWarning(
                        "Tool Controller Vertical Rapid Values are unset" + "\n"
                    )

        commands = PathUtils.getPathWithPlacement(pathobj).Commands
        for index, c in enumerate(commands):

            outstring = []
            outsuffix = []
            command = c.Name
            if index + 1 == len(commands):
                nextcommand = ""
            else:
                nextcommand = commands[index + 1].Name

            if adaptiveOp and c.Name in ["G0", "G00"]:
                if opHorizRapid and opVertRapid:
                    command = "G1"
                else:
                    outstring.append("(TOOL CONTROLLER RAPID VALUES ARE UNSET)" + "\n")

            # suppress moves in fixture selection
            if pathobj.Label == "Fixture":
                if command == "G0":
                    continue

            # If tool a tap, we will thread tap, so stop the spindle
            # for now as there is no point in starting it to stop it
            # in the G74/G84 operation after S29.  This only trigger
            # when pathobj is a ToolController.
            if command == "M03" or command == "M3":
                if (
                    hasattr(pathobj, "Tool")
                    and getattr(pathobj.Tool, "ShapeType", "").lower() == "tap"
                ):
                    tapSpeed = int(pathobj.SpindleSpeed)
                    continue

            # Handle thread tapping cycles.  Uses rigid tapping.
            # This only trigger when pathobj is a Operation.
            if command == "G74" or command == "G84":
                pitch_mm = float(c.Parameters["F"])
                # Convert pitch to inches if needed
                if UNITS == "G20":  # imperial
                    pitch = pitch_mm / 25.4
                else:
                    pitch = pitch_mm
                paramstring = ""
                for param in ["X", "Y"]:
                    if param in c.Parameters:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                            paramstring += (
                                " "
                                + param
                                + format(
                                    float(pos.getValueAs(UNIT_FORMAT)),
                                    precision_string,
                                )
                            )
                if paramstring != "":
                    out += linenumber() + "G00" + paramstring + "\n"

                if "S" in c.Parameters:
                    tapSpeed = int(c.Parameters["S"])
                out += "M29 S" + str(tapSpeed) + "\n"

                for param in ["Z", "R"]:
                    if param in c.Parameters:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                            paramstring += (
                                " "
                                + param
                                + format(
                                    float(pos.getValueAs(UNIT_FORMAT)),
                                    precision_string,
                                )
                            )

                # Calculate feed rate as distance per minute
                if tapSpeed is not None:
                    feed_rate = pitch * tapSpeed
                    speed = Units.Quantity(feed_rate, UNIT_SPEED_FORMAT)
                    paramstring += " F" + format(
                        float(speed.getValueAs(UNIT_SPEED_FORMAT)), precision_string
                    )
                else:
                    # No spindle speed found, output pitch as F
                    paramstring += " F" + format(pitch, precision_string)

                # P is the dwell time in seconds at the bottom of the thread
                # Q is the peck depth of the threading operation
                for param in ["P", "Q"]:
                    if param in c.Parameters:
                        value = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                        paramstring += (
                            " "
                            + param
                            + format(
                                float(value.getValueAs(UNIT_FORMAT)),
                                precision_string,
                            )
                        )

                out += linenumber() + command + paramstring + "\n"
                out += linenumber() + "G80\n"  # End tapping cycle
                continue

            outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            # suppress a G80 between two identical command
            if command == "G80" and lastcommand == nextcommand:
                continue

            if c.Name.startswith("(") and not OUTPUT_COMMENTS:  # command is a comment
                continue

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F" and (
                        currLocation[param] != c.Parameters[param] or OUTPUT_DOUBLES
                    ):
                        if c.Name not in [
                            "G0",
                            "G00",
                        ]:  # fanuc doesn't use rapid speeds
                            speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                                outstring.append(
                                    param
                                    + format(
                                        float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                            else:
                                continue
                    elif param == "T":
                        outstring.append(param + str(int(c.Parameters["T"])))
                    elif param == "H":
                        outstring.append(param + str(int(c.Parameters["H"])))
                    elif param == "D":
                        outstring.append(param + str(int(c.Parameters["D"])))
                    elif param == "S":
                        outstring.append(param + str(int(c.Parameters["S"])))
                    else:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and currLocation[param] == c.Parameters[param]
                            and (
                                command not in DRILL_OPERATION
                                or (command in DRILL_OPERATION and param not in DRILL_PARAM_REQ)
                            )
                        ):
                            continue
                        else:
                            pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                            outstring.append(
                                param + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                            )

            if adaptiveOp and c.Name in ["G0", "G00"]:
                if opHorizRapid and opVertRapid:
                    if "Z" not in c.Parameters:
                        outstring.append(
                            "F"
                            + format(
                                float(opHorizRapid.getValueAs(UNIT_SPEED_FORMAT)),
                                precision_string,
                            )
                        )
                    else:
                        outstring.append(
                            "F"
                            + format(
                                float(opVertRapid.getValueAs(UNIT_SPEED_FORMAT)),
                                precision_string,
                            )
                        )

            # store the latest command
            lastcommand = command
            currLocation.update(c.Parameters)

            # Check for Tool Change:
            if command == "M6":
                # stop the spindle
                out += linenumber() + "M05\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

                # add height offset
                if USE_TLO:
                    outsuffix.append("G91 G0 G43 G54 Z-[#[2000+#4120]] H#4120")
                    outsuffix.append("G90")

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                out += COMMAND_SPACE.join(outstring).upper()
                out = out.strip() + "\n"
            if len(outsuffix) >= 1:
                for line in outsuffix:
                    out += linenumber() + line + "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
from unittest.mock import patch, MagicMock

from importXLSX import FormulaTranslator, getText, handleStrings, open


class TestFormulaTranslator(unittest.TestCase):
    def test_translate_expressions(self):
        # With
        formulas_and_expressions = [
            ("1=2", "1==2"),
            ("a<b", "a<b"),
            ("a>b", "a>b"),
            ("1<>2", "1!=2"),
            ("a>=b", "a>=b"),
            ("a<=b", "a<=b"),
            ("a!b", "a.b"),
            ("a+b", "a+b"),
            ("a-b", "a-b"),
            ("a*b", "a*b"),
            ("a/b", "a/b"),
            ("a^b", "a^b"),
            ("c*(a+b)", "c*(a+b)"),
            ("IF(a=b, c, d)", "(a==b? c: d)"),
            ("ABS(a)", "abs(a)"),
            ("ACOS(a)", "pi/180deg*acos(a)"),
            ("ASIN(a)", "pi/180deg*asin(a)"),
            ("ATAN(a)", "pi/180deg*atan(a)"),
            ("ATAN2(a)", "pi/180deg*atan2(a)"),
            ("COS(a)", "cos(1rad*(a))"),
            ("COSH(a)", "cosh(1rad*(a))"),
            ("EXP(a)", "exp(a)"),
            ("LOG(n)", "log(n)"),
            ("LOG10(n)", "log10(n)"),
            ("MOD(n,d)", "mod(n,d)"),
            ("POWER(n,p)", "pow(n,p)"),
            ("SIN(a)", "sin(1rad*(a))"),
            ("SINH(a)", "sinh(1rad*(a))"),
            ("SQRT(a)", "sqrt(a)"),
            ("TAN(a)", "tan(1rad*(a))"),
            ("TANH(a)", "tanh(1rad*(a))"),
            ("AVERAGE(a,b,c)", "average(a,b,c)"),
            ("COUNT(a,b,c)", "count(a,b,c)"),
            ("MAX(a,b,c)", "max(a,b,c)"),
            ("MIN(a,b,c)", "min(a,b,c)"),
            ("STDEVA(a,b,c)", "stddev(a,b,c)"),
            ("SUM(a,b,c)", "sum(a,b,c)"),
            ("PI", "pi"),
            ("_xlfn.CEILING.MATH(a)", "ceil(a)"),
            ("_xlfn.FLOOR.MATH(a)", "floor(a)"),
        ]

        # When
        result = []
        for formula, _ in formulas_and_expressions:
            translator = FormulaTranslator()
            result.append(translator.translateForm(formula))

        # Then
        expected = [f"={expression}" for _, expression in formulas_and_expressions]
        self.assertListEqual(expected, result)

    def test_translate_multi_character_branching_operators(self):
        # With
        formulas_and_expressions = [
            ("1<>2", "1!=2"),
            ("a>=b", "a>=b"),
            ("a<=b", "a<=b"),
        ]

        # When
        result = []
        for formula, _ in formulas_and_expressions:
            translator = FormulaTranslator()
            result.append(translator.translateForm(formula))

        # Then
        expected = [f"={expression}" for _, expression in formulas_and_expressions]
        self.assertListEqual(expected, result)

    def test_translate_nested_expression(self):
        # With
        formulas_and_expressions = [
            (
                "IF(a<b+c,0.1+SIN(0.5),0.3+MAX(COS(0.2),SIN(0.1)))",
                "(a<b+c?0.1+sin(1rad*(0.5)):0.3+max(cos(1rad*(0.2)),sin(1rad*(0.1))))",
            ),
        ]

        # When
        result = []
        for formula, _ in formulas_and_expressions:
            translator = FormulaTranslator()
            result.append(translator.translateForm(formula))

        # Then
        expected = [f"={expression}" for _, expression in formulas_and_expressions]
        self.assertListEqual(expected, result)



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

import re

import addonmanager_freecad_interface as fci

from PySideWrapper import QtCore, QtGui, QtWidgets

from typing import Optional


class WidgetReadmeBrowser(QtWidgets.QTextBrowser):
    """A QTextBrowser widget that emits signals for each requested image resource, allowing an external controller
    to load and re-deliver those images. Once all resources have been re-delivered, the original data is redisplayed
    with the images in-line. Call setUrl prior to calling setMarkdown or setHtml to ensure URLs are resolved
    correctly."""

    load_resource = QtCore.Signal(str)  # Str is a URL to a resource
    follow_link = QtCore.Signal(str)  # Str is a URL to another page

    def __init__(self, parent: QtWidgets.QWidget = None):
        super().__init__(parent)
        self.image_map = {}
        self.url = ""
        self.stop = False
        self.setOpenExternalLinks(True)

    def setUrl(self, url: str):
        """Set the base URL of the page. Used to resolve relative URLs in the page source."""
        self.url = url

    def setMarkdown(self, md: str):
        """Provides an optional fallback to the markdown library for older versions of Qt (prior to 5.15) that did not
        have native markdown support. Lacking that, plaintext is displayed."""
        geometry = self.geometry()
        if hasattr(super(), "setMarkdown"):

            super().setMarkdown(self._clean_markdown(md))
        else:
            try:
                import markdown

                html = markdown.markdown(md)
                self.setHtml(html)
            except ImportError:
                self.setText(md)
                fci.Console.Warning(
                    "Qt < 5.15 and no `import markdown` -- falling back to plain text display\n"
                )
        self.setGeometry(geometry)

    def _clean_markdown(self, md: str):
        # Remove some HTML tags (for now just img and br, which are the most common offenders that break rendering)
        br_re = re.compile(r"<br\s*/?>", re.IGNORECASE)
        comment_re = re.compile(r"<!--.*?-->", re.DOTALL)
        img_re = re.compile(
            r'<img\s+[^>]*src=["\'](?P<src>[^"\']+)["\'][^>]*'
            r'(alt=["\'](?P<alt>[^"\']*)["\'])?[^>]*/?>',
            re.IGNORECASE,
        )

        # Replace html images to markdown
        def _markdown_img(m):
            src = m.group("src")
            alt = m.group("alt") or ""
            return f"![{alt}]({src})"

        cleaned = br_re.sub(r"\n", md)
        cleaned = comment_re.sub("", cleaned)
        cleaned = img_re.sub(_markdown_img, cleaned)

        return cleaned

    def set_resource(self, resource_url: str, image: Optional[QtGui.QImage]):
        """Once a resource has been fetched (or the fetch has failed), this method should be used to inform the widget
        that the resource has been loaded. Note that the incoming image is scaled to 97% of the widget width if it is
        larger than that."""
        self.image_map[resource_url] = self._ensure_appropriate_width(image)

    def loadResource(self, resource_type: int, name: QtCore.QUrl) -> object:
        """Callback for resource loading. Called automatically by underlying Qt
        code when external resources are needed for rendering. In particular,
        here it is used to download and cache (in RAM) the images needed for the
        README and Wiki pages."""
        if resource_type == QtGui.QTextDocument.ImageResource and not self.stop:
            full_url = self._create_full_url(name.toString())
            if full_url not in self.image_map:
                self.load_resource.emit(full_url)
                self.image_map[full_url] = None
            return self.image_map[full_url]
        elif resource_type == QtGui.QTextDocument.MarkdownResource:
            self.follow_link.emit(name.toString())
            return self.toMarkdown()
        elif resource_type == QtGui.QTextDocument.HtmlResource:
            self.follow_link.emit(name.toString())
            return self.toHtml()
        return super().loadResource(resource_type, name)

    def _ensure_appropriate_width(self, image: QtGui.QImage) -> QtGui.QImage:
        ninety_seven_percent = self.width() * 0.97
        if image.width() < ninety_seven_percent:
            return image
        return image.scaledToWidth(ninety_seven_percent)

    def _create_full_url(self, url: str) -> str:
        if url.startswith("http"):
            return url
        if not self.url:
            return url
        lhs, slash, _ = self.url.rpartition("/")
        return lhs + slash + url



# ***************************************************************************
# *   Copyright (c) 2020 Markus Hovorka <m.hovorka@live.de>                 *
# *   Copyright (c) 2022 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM Electricforce Elmer writer"
__author__ = "Markus Hovorka, Uwe Stöhr"
__url__ = "https://www.freecad.org"

## \addtogroup FEM
#  @{

from .. import sifio
from . import electricforce


class EFwriter:

    def __init__(self, writer, solver):
        self.write = writer
        self.solver = solver

    def getElectricforceSolver(self, equation):
        # check if we need to update the equation
        self._updateElectricforceSolver(equation)
        # output the equation parameters
        s = self.write.createEmptySolver()
        s["Equation"] = "Electric Force"  # equation.Name
        s["Procedure"] = sifio.FileAttr("ElectricForce/StatElecForce")
        s["Exec Solver"] = equation.ExecSolver
        s["Stabilize"] = equation.Stabilize
        return s

    def _updateElectricforceSolver(self, equation):
        # updates older Electricforce equations
        if not hasattr(equation, "ExecSolver"):
            equation.addProperty(
                "App::PropertyEnumeration",
                "ExecSolver",
                "Electric Force",
                (
                    "That solver is only executed after solution converged\n"
                    "To execute always, change to 'Always'"
                ),
                locked=True,
            )
            equation.ExecSolver = electricforce.SOLVER_EXEC_METHODS
            equation.ExecSolver = "After Timestep"


##  @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# -*- coding: utf-8 -*
# ***************************************************************************
# *   Copyright (c) 2018 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# *   Bilinear interpolation code modified heavily from the interpolation   *
# *   library https://github.com/pmav99/interpolation                      *
# *   Copyright (c) 2013 by Panagiotis Mavrogiorgos                         *
# *                                                                         *
# ***************************************************************************
import FreeCAD
import FreeCADGui
import Path
import PathScripts.PathUtils as PathUtils
import Path.Dressup.Utils as PathDressup

from PySide import QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

Part = LazyLoader("Part", globals(), "Part")

"""Z Depth Correction Dressup.  This dressup takes a probe file as input and does bilinear interpolation of the Zdepths to correct for a surface which is not parallel to the milling table/bed.  The probe file should conform to the format specified by the linuxcnc G38 probe logging: 9-number coordinate consisting of XYZABCUVW http://linuxcnc.org/docs/html/gcode/g-code.html#gcode:g38
"""

LOGLEVEL = False

LOG_MODULE = Path.Log.thisModule()

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


class ObjectDressup:
    def __init__(self, obj):
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Path",
            QT_TRANSLATE_NOOP("App::Property", "The base toolpath to modify"),
        )
        obj.addProperty(
            "App::PropertyFile",
            "probefile",
            "ProbeData",
            QT_TRANSLATE_NOOP("App::Property", "The point file from the surface probing."),
        )
        obj.Proxy = self
        obj.addProperty("Part::PropertyPartShape", "interpSurface", "Path")
        obj.addProperty(
            "App::PropertyDistance",
            "ArcInterpolate",
            "Interpolate",
            QT_TRANSLATE_NOOP("App::Property", "Deflection distance for arc interpolation"),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "SegInterpolate",
            "Interpolate",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "break segments into smaller segments of this length.",
            ),
        )
        obj.ArcInterpolate = 0.1
        obj.SegInterpolate = 1.0

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onChanged(self, obj, prop):
        if str(prop) == "probefile":
            self._loadFile(obj, obj.probefile)
        if prop == "Path" and obj.ViewObject:
            obj.ViewObject.signalChangeIcon()

    def _bilinearInterpolate(self, surface, x, y):
        p1 = FreeCAD.Vector(x, y, 100.0)
        p2 = FreeCAD.Vector(x, y, -100.0)

        vertical_line = Part.Line(p1, p2)
        points, curves = vertical_line.intersectCS(surface)
        return points[0].Z

    def _loadFile(self, obj, filename):
        if not filename:
            return

        f1 = open(filename, "r")

        try:
            pointlist = []
            for line in f1.readlines():
                if line == "\n":
                    continue
                w = line.split()
                xval = round(float(w[0]), 2)
                yval = round(float(w[1]), 2)
                zval = round(float(w[2]), 2)

                pointlist.append([xval, yval, zval])
            Path.Log.debug(pointlist)

            cols = list(zip(*pointlist))
            Path.Log.debug("cols: {}".format(cols))
            yindex = list(sorted(set(cols[1])))
            Path.Log.debug("yindex: {}".format(yindex))

            array = []
            for y in yindex:
                points = sorted([p for p in pointlist if p[1] == y])
                inner = []
                for p in points:
                    inner.append(FreeCAD.Vector(p[0], p[1], p[2]))
                array.append(inner)

            intSurf = Part.BSplineSurface()
            intSurf.interpolate(array)

            obj.interpSurface = intSurf.toShape()
        except Exception:
            raise ValueError("File does not contain appropriate point data")

    def execute(self, obj):
        sampleD = obj.SegInterpolate.Value
        curveD = obj.ArcInterpolate.Value

        if obj.interpSurface.isNull():  # No valid probe data.  return unchanged path
            obj.Path = PathUtils.getPathWithPlacement(obj.Base)
            return

        surface = obj.interpSurface.toNurbs().Faces[0].Surface

        if obj.Base:
            if obj.Base.isDerivedFrom("Path::Feature"):
                if obj.Base.Path:
                    path = PathUtils.getPathWithPlacement(obj.Base)
                    if path.Commands:
                        pathlist = path.Commands

                        newcommandlist = []
                        currLocation = {"X": 0, "Y": 0, "Z": 0, "F": 0}

                        for c in pathlist:
                            Path.Log.debug(c)
                            Path.Log.debug("     curLoc:{}".format(currLocation))
                            newparams = dict(c.Parameters)
                            zval = newparams.get("Z", currLocation["Z"])
                            if c.Name in Path.Geom.CmdMoveStraight + Path.Geom.CmdMoveArc:
                                curVec = FreeCAD.Vector(
                                    currLocation["X"],
                                    currLocation["Y"],
                                    currLocation["Z"],
                                )
                                arcwire = Path.Geom.edgeForCmd(c, curVec)
                                if arcwire is None:
                                    continue
                                if c.Name in Path.Geom.CmdMoveArc:
                                    pointlist = arcwire.discretize(Deflection=curveD)
                                else:
                                    disc_number = int(arcwire.Length / sampleD)
                                    if disc_number > 1:
                                        pointlist = arcwire.discretize(
                                            Number=int(arcwire.Length / sampleD)
                                        )
                                    else:
                                        pointlist = [v.Point for v in arcwire.Vertexes]
                                for point in pointlist:
                                    offset = self._bilinearInterpolate(surface, point.x, point.y)

                                    commandparams = {
                                        "X": point.x,
                                        "Y": point.y,
                                        "Z": point.z + offset,
                                    }
                                    if "F" in newparams.keys():
                                        commandparams["F"] = newparams["F"]
                                    newcommand = Path.Command("G1", commandparams)

                                    newcommandlist.append(newcommand)
                                    currLocation.update(newcommand.Parameters)
                                    currLocation["Z"] = zval

                            else:
                                # Non Feed Command
                                newcommandlist.append(c)
                                currLocation.update(c.Parameters)
                        path = Path.Path(newcommandlist)
                        obj.Path = path


class TaskPanel:
    def __init__(self, obj):
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/ZCorrectEdit.ui")
        FreeCAD.ActiveDocument.openTransaction("Edit Z Correction Dress-up")
        self.interpshape = FreeCAD.ActiveDocument.addObject("Part::Feature", "InterpolationSurface")
        self.interpshape.Shape = obj.interpSurface
        self.interpshape.ViewObject.Transparency = 60
        self.interpshape.ViewObject.ShapeColor = (1.00000, 1.00000, 0.01961)
        self.interpshape.ViewObject.Selectable = False
        stock = PathUtils.findParentJob(obj).Stock
        self.interpshape.Placement.Base.z = stock.Shape.BoundBox.ZMax

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()

    def accept(self):
        self.getFields()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.removeObject(self.interpshape.Name)
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCAD.ActiveDocument.recompute()

    def getFields(self):
        self.obj.Proxy.execute(self.obj)

    def updateUI(self):
        if Path.Log.getLevel(LOG_MODULE) == Path.Log.Level.DEBUG:
            for obj in FreeCAD.ActiveDocument.Objects:
                if obj.Name.startswith("Shape"):
                    FreeCAD.ActiveDocument.removeObject(obj.Name)
            print("object name %s" % self.obj.Name)
            if hasattr(self.obj.Proxy, "shapes"):
                Path.Log.info("showing shapes attribute")
                for shapes in self.obj.Proxy.shapes.itervalues():
                    for shape in shapes:
                        Part.show(shape)
            else:
                Path.Log.info("no shapes attribute found")

    def updateModel(self):
        self.getFields()
        self.updateUI()
        FreeCAD.ActiveDocument.recompute()

    def setFields(self):
        self.form.ProbePointFileName.setText(self.obj.probefile)

        self.updateUI()

    def open(self):
        pass

    def setupUi(self):
        self.setFields()
        # now that the form is filled, setup the signal handlers
        self.form.ProbePointFileName.editingFinished.connect(self.updateModel)
        self.form.SetProbePointFileName.clicked.connect(self.SetProbePointFileName)

    def SetProbePointFileName(self):
        filename = QtGui.QFileDialog.getOpenFileName(
            self.form,
            translate("CAM_Probe", "Select Probe Point File"),
            None,
            translate("CAM_Probe", "All Files (*.*)"),
        )
        if filename and filename[0]:
            self.obj.probefile = str(filename[0])
            self.setFields()


class ViewProviderDressup:
    def __init__(self, vobj):
        vobj.Proxy = self

    def attach(self, vobj):
        self.obj = vobj.Object
        if self.obj and self.obj.Base:
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
        return

    def claimChildren(self):
        return [self.obj.Base]

    def setEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        panel = TaskPanel(vobj.Object)
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onDelete(self, arg1=None, arg2=None):
        """this makes sure that the base operation is added back to the project and visible"""
        FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
        job = PathUtils.findParentJob(arg1.Object)
        job.Proxy.addOperation(arg1.Object.Base)
        arg1.Object.Base = None
        return True

    def getIcon(self):
        if getattr(PathDressup.baseOp(self.obj), "Active", True):
            return ":/icons/CAM_Dressup.svg"
        else:
            return ":/icons/CAM_OpActive.svg"


class CommandPathDressup:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupZCorrect", "Z Depth Correction"),
            "Accel": "",
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupZCorrect", "Corrects Z depth using a probe map"
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):
        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(translate("CAM_Dressup", "Select one toolpath object\n"))
            return
        if not selection[0].isDerivedFrom("Path::Feature"):
            FreeCAD.Console.PrintError(
                translate("CAM_Dressup", "The selected object is not a toolpath\n")
            )
            return
        if selection[0].isDerivedFrom("Path::FeatureCompoundPython"):
            FreeCAD.Console.PrintError(translate("CAM_Dressup", "Select a toolpath object"))
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.ZCorrect")
        FreeCADGui.addModule("PathScripts.PathUtils")
        FreeCADGui.doCommand(
            'obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", "ZCorrectDressup")'
        )
        FreeCADGui.doCommand("Path.Dressup.Gui.ZCorrect.ObjectDressup(obj)")
        FreeCADGui.doCommand("obj.Base = FreeCAD.ActiveDocument." + selection[0].Name)
        FreeCADGui.doCommand("Path.Dressup.Gui.ZCorrect.ViewProviderDressup(obj.ViewObject)")
        FreeCADGui.doCommand("PathScripts.PathUtils.addToJob(obj)")
        FreeCADGui.doCommand("Gui.ActiveDocument.getObject(obj.Base.Name).Visibility = False")
        FreeCADGui.doCommand("obj.ViewObject.Document.setEdit(obj.ViewObject, 0)")
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_DressupZCorrect", CommandPathDressup())

FreeCAD.Console.PrintLog("Loading PathDressup… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Base.Util as PathUtil
import Path.Base.SetupSheetOpPrototype as PathSetupSheetOpPrototype
from PySide.QtCore import QT_TRANSLATE_NOOP

__title__ = "Setup Sheet for a Job."
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "A container for all default values and job specific configuration values."

_RegisteredOps: dict = {}


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class Template:

    HorizRapid = "HorizRapid"
    VertRapid = "VertRapid"
    CoolantMode = "CoolantMode"
    SafeHeightOffset = "SafeHeightOffset"
    SafeHeightExpression = "SafeHeightExpression"
    ClearanceHeightOffset = "ClearanceHeightOffset"
    ClearanceHeightExpression = "ClearanceHeightExpression"
    StartDepthExpression = "StartDepthExpression"
    FinalDepthExpression = "FinalDepthExpression"
    StepDownExpression = "StepDownExpression"
    Fixtures = "Fixtures"
    OrderOutputBy = "OrderOutputBy"
    SplitOutput = "SplitOutput"

    All = [
        HorizRapid,
        VertRapid,
        CoolantMode,
        SafeHeightOffset,
        SafeHeightExpression,
        ClearanceHeightOffset,
        ClearanceHeightExpression,
        StartDepthExpression,
        FinalDepthExpression,
        StepDownExpression,
    ]


def _traverseTemplateAttributes(attrs, codec):
    Path.Log.debug(attrs)
    coded = {}
    for key, value in attrs.items():
        if type(value) == dict:
            Path.Log.debug("%s is a dict" % key)
            coded[key] = _traverseTemplateAttributes(value, codec)
        elif type(value) == list:
            Path.Log.debug("%s is a list" % key)
            coded[key] = [_traverseTemplateAttributes(attr, codec) for attr in value]
        elif isinstance(value, str):
            Path.Log.debug("%s is a string" % key)
            coded[key] = codec(value)
        else:
            Path.Log.debug("%s is %s" % (key, type(value)))
            coded[key] = value
    return coded


class SetupSheet:
    """Property container object used by a Job to hold global reference values."""

    TemplateReference = "${SetupSheet}"

    DefaultSafeHeightOffset = "3 mm"
    DefaultClearanceHeightOffset = "5 mm"
    DefaultSafeHeightExpression = "OpStockZMax+${SetupSheet}.SafeHeightOffset"
    DefaultClearanceHeightExpression = "OpStockZMax+${SetupSheet}.ClearanceHeightOffset"

    DefaultStartDepthExpression = "OpStartDepth"
    DefaultFinalDepthExpression = "OpFinalDepth"
    DefaultStepDownExpression = "OpToolDiameter"

    DefaultCoolantModes = ["None", "Flood", "Mist"]

    def __init__(self, obj):
        self.obj = obj
        obj.addProperty(
            "App::PropertySpeed",
            "VertRapid",
            "ToolController",
            QT_TRANSLATE_NOOP("App::Property", "Default speed for horizontal rapid moves."),
        )
        obj.addProperty(
            "App::PropertySpeed",
            "HorizRapid",
            "ToolController",
            QT_TRANSLATE_NOOP("App::Property", "Default speed for vertical rapid moves."),
        )
        obj.addProperty(
            "App::PropertyStringList",
            "CoolantModes",
            "CoolantMode",
            QT_TRANSLATE_NOOP("App::Property", "Coolant Modes"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "CoolantMode",
            "CoolantMode",
            QT_TRANSLATE_NOOP("App::Property", "Default coolant mode."),
        )
        obj.addProperty(
            "App::PropertyLength",
            "SafeHeightOffset",
            "OperationHeights",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The usage of this field depends on SafeHeightExpression - by default its value is added to the start depth and used for the safe height of an operation.",
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "SafeHeightExpression",
            "OperationHeights",
            QT_TRANSLATE_NOOP("App::Property", "Expression for the safe height of new operations."),
        )
        obj.addProperty(
            "App::PropertyLength",
            "ClearanceHeightOffset",
            "OperationHeights",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The usage of this field depends on ClearanceHeightExpression - by default is value is added to the start depth and used for the clearance height of an operation.",
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "ClearanceHeightExpression",
            "OperationHeights",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Expression for the clearance height of new operations.",
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "StartDepthExpression",
            "OperationDepths",
            QT_TRANSLATE_NOOP(
                "App::Property", "Expression used for the start depth of new operations."
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "FinalDepthExpression",
            "OperationDepths",
            QT_TRANSLATE_NOOP(
                "App::Property", "Expression used for the final depth of new operations."
            ),
        )
        obj.addProperty(
            "App::PropertyString",
            "StepDownExpression",
            "OperationDepths",
            QT_TRANSLATE_NOOP("App::Property", "Expression used for step down of new operations."),
        )

        obj.SafeHeightOffset = self.decodeAttributeString(self.DefaultSafeHeightOffset)
        obj.ClearanceHeightOffset = self.decodeAttributeString(self.DefaultClearanceHeightOffset)
        obj.SafeHeightExpression = self.decodeAttributeString(self.DefaultSafeHeightExpression)
        obj.ClearanceHeightExpression = self.decodeAttributeString(
            self.DefaultClearanceHeightExpression
        )

        obj.StartDepthExpression = self.decodeAttributeString(self.DefaultStartDepthExpression)
        obj.FinalDepthExpression = self.decodeAttributeString(self.DefaultFinalDepthExpression)
        obj.StepDownExpression = self.decodeAttributeString(self.DefaultStepDownExpression)

        obj.CoolantModes = self.DefaultCoolantModes
        obj.CoolantMode = self.DefaultCoolantModes

        obj.Proxy = self

    def dumps(self):
        return None

    def loads(self, state):
        for obj in FreeCAD.ActiveDocument.Objects:
            if hasattr(obj, "Proxy") and obj.Proxy == self:
                self.obj = obj
                break
        return None

    def hasDefaultToolRapids(self):
        return Path.Geom.isRoughly(self.obj.VertRapid.Value, 0) and Path.Geom.isRoughly(
            self.obj.HorizRapid.Value, 0
        )

    def hasDefaultOperationHeights(self):
        if (
            self.obj.SafeHeightOffset.UserString
            != FreeCAD.Units.Quantity(self.DefaultSafeHeightOffset).UserString
        ):
            return False
        if (
            self.obj.ClearanceHeightOffset.UserString
            != FreeCAD.Units.Quantity(self.DefaultClearanceHeightOffset).UserString
        ):
            return False
        if self.obj.SafeHeightExpression != self.decodeAttributeString(
            self.DefaultSafeHeightExpression
        ):
            return False
        if self.obj.ClearanceHeightExpression != self.decodeAttributeString(
            self.DefaultClearanceHeightExpression
        ):
            return False
        return True

    def hasDefaultOperationDepths(self):
        if self.obj.StartDepthExpression != self.DefaultStartDepthExpression:
            return False
        if self.obj.FinalDepthExpression != self.DefaultFinalDepthExpression:
            return False
        if self.obj.StepDownExpression != self.DefaultStepDownExpression:
            return False
        return True

    def hasDefaultCoolantMode(self):
        return self.obj.CoolantMode == "None"

    def setFromTemplate(self, attrs):
        """setFromTemplate(attrs) ... sets the default values from the given dictionary."""
        for name in Template.All:
            if attrs.get(name) is not None:
                setattr(self.obj, name, attrs[name])

        for opName, op in _RegisteredOps.items():
            opSetting = attrs.get(opName)
            if opSetting is not None:
                prototype = op.prototype(opName)
                for propName in op.properties():
                    value = opSetting.get(propName)
                    if value is not None:
                        prop = prototype.getProperty(propName)
                        propertyName = OpPropertyName(opName, propName)
                        propertyGroup = OpPropertyGroup(opName)
                        prop.setupProperty(
                            self.obj,
                            propertyName,
                            propertyGroup,
                            prop.valueFromString(value),
                        )

    def templateAttributes(
        self,
        includeRapids=True,
        includeCoolantMode=True,
        includeHeights=True,
        includeDepths=True,
        includeOps=None,
    ):
        """templateAttributes(includeRapids, includeHeights, includeDepths) ... answers a dictionary with the default values."""
        attrs = {}

        if includeRapids:
            attrs[Template.VertRapid] = self.obj.VertRapid.UserString
            attrs[Template.HorizRapid] = self.obj.HorizRapid.UserString

        if includeCoolantMode:
            attrs[Template.CoolantMode] = self.obj.CoolantMode

        if includeHeights:
            attrs[Template.SafeHeightOffset] = self.obj.SafeHeightOffset.UserString
            attrs[Template.SafeHeightExpression] = self.obj.SafeHeightExpression
            attrs[Template.ClearanceHeightOffset] = self.obj.ClearanceHeightOffset.UserString
            attrs[Template.ClearanceHeightExpression] = self.obj.ClearanceHeightExpression

        if includeDepths:
            attrs[Template.StartDepthExpression] = self.obj.StartDepthExpression
            attrs[Template.FinalDepthExpression] = self.obj.FinalDepthExpression
            attrs[Template.StepDownExpression] = self.obj.StepDownExpression

        if includeOps:
            for opName in includeOps:
                settings = {}
                op = _RegisteredOps[opName]
                for propName in op.properties():
                    prop = OpPropertyName(opName, propName)
                    if hasattr(self.obj, prop):
                        settings[propName] = PathUtil.getPropertyValueString(self.obj, prop)
                attrs[opName] = settings

        return attrs

    def expressionReference(self):
        """expressionReference() ... returns the string to be used in expressions"""
        # Using the Name here and not the Label (both would be valid) because the Name 'fails early'.
        #
        # If there is a Name/Label conflict and an expression is bound to the Name we'll get an error
        # on creation (Property not found). Not good, but at least there's some indication that
        # something's afoul.
        #
        # If the expression is based on the Label everything works out nicely - until the document is
        # saved and loaded from disk. The Labels change in order to avoid the Name/Label conflict
        # but the expression stays the same. If the user's lucky the expression is broken because the
        # conflicting object doesn't have the properties reference by the expressions. If the user is
        # not so lucky those properties also exist in the other object, there is no indication that
        # anything is wrong but the expressions will substitute the values from the wrong object.
        #
        # I prefer the question: "why do I get this error when I create ..." over "my cnc machine just
        # rammed it's tool head into the table ..." or even "I saved my file and now it's corrupt..."
        #
        # https://forum.freecad.org/viewtopic.php?f=10&t=24839
        # https://forum.freecad.org/viewtopic.php?f=10&t=24845
        return self.obj.Name

    def encodeAttributeString(self, attr):
        """encodeAttributeString(attr) ... return the encoded string of a template attribute."""
        return str(attr.replace(self.expressionReference(), self.TemplateReference))

    def decodeAttributeString(self, attr):
        """decodeAttributeString(attr) ... return the decoded string of a template attribute."""
        return str(attr.replace(self.TemplateReference, self.expressionReference()))

    def encodeTemplateAttributes(self, attrs):
        """encodeTemplateAttributes(attrs) ... return a dictionary with all values encoded."""
        return _traverseTemplateAttributes(attrs, self.encodeAttributeString)

    def decodeTemplateAttributes(self, attrs):
        """decodeTemplateAttributes(attrs) ... expand template attributes to reference the receiver where applicable."""
        return _traverseTemplateAttributes(attrs, self.decodeAttributeString)

    def operationsWithSettings(self):
        """operationsWithSettings() ... returns a list of operations which currently have some settings defined."""
        ops = []
        for name, value in _RegisteredOps.items():
            for prop in value.registeredPropertyNames(name):
                if hasattr(self.obj, prop):
                    ops.append(name)
                    break
        return list(sorted(ops))

    def setOperationProperties(self, obj, opName):
        Path.Log.track(obj.Label, opName)
        try:
            op = _RegisteredOps[opName]
            for prop in op.properties():
                propName = OpPropertyName(opName, prop)
                if hasattr(self.obj, propName):
                    setattr(obj, prop, getattr(self.obj, propName))
        except Exception:
            Path.Log.info("SetupSheet has no support for {}".format(opName))

    def onDocumentRestored(self, obj):

        if not hasattr(obj, "CoolantModes"):
            obj.addProperty(
                "App::PropertyStringList",
                "CoolantModes",
                "CoolantMode",
                QT_TRANSLATE_NOOP("App::Property", "Coolant Modes"),
            )
            obj.CoolantModes = self.DefaultCoolantModes

        if not hasattr(obj, "CoolantMode"):
            obj.addProperty(
                "App::PropertyEnumeration",
                "CoolantMode",
                "CoolantMode",
                QT_TRANSLATE_NOOP("App::Property", "Default coolant mode."),
            )
            obj.CoolantMode = self.DefaultCoolantModes


def Create(name="SetupSheet"):
    obj = FreeCAD.ActiveDocument.addObject("App::FeaturePython", name)
    obj.Proxy = SetupSheet(obj)
    return obj


class _RegisteredOp(object):
    def __init__(self, factory, properties):
        self.factory = factory
        self.properties = properties

    def registeredPropertyNames(self, name):
        return [OpPropertyName(name, prop) for prop in self.properties()]

    def prototype(self, name):
        ptt = PathSetupSheetOpPrototype.OpPrototype(name)
        self.factory("OpPrototype.%s" % name, ptt)
        return ptt


def RegisterOperation(name, objFactory, setupProperties):
    global _RegisteredOps
    _RegisteredOps[name] = _RegisteredOp(objFactory, setupProperties)


def OpNamePrefix(name):
    return name.replace("Path", "").replace(" ", "").replace("_", "")


def OpPropertyName(opName, propName):
    return "{}{}".format(OpNamePrefix(opName), propName)


def OpPropertyGroup(opName):
    return "Op {}".format(opName)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2018-25 Paul Lee <paullee0@gmail.com>                   *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import ArchWindow
from PySide.QtCore import QT_TRANSLATE_NOOP


class ArchSketchObject:
    def __init__(self, obj):
        pass


class ArchSketch(ArchSketchObject):
    def __init__(self, obj):
        pass

    def setPropertiesLinkCommon(self, orgFp, linkFp=None, mode=None):
        if linkFp:
            fp = linkFp
        else:
            fp = orgFp
        prop = fp.PropertiesList
        if not isinstance(fp.getLinkedObject().Proxy, ArchWindow._Window):
            pass
        else:
            if "Hosts" not in prop:
                # inherited properties of Link are not in PropertiesList:
                old_hosts = getattr(fp, "Hosts", [])
                fp.addProperty(
                    "App::PropertyLinkList",
                    "Hosts",
                    "Window",
                    QT_TRANSLATE_NOOP("App::Property", "The objects that host this window"),
                    locked=True,
                )
                fp.Hosts = old_hosts
                for host in old_hosts:
                    host.touch()
                # Arch Window's code


# from ArchSketchObjectExt import ArchSketch  # Doesn't work



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from PySide import QtGui, QtCore
from .shapewidget import ShapeWidget


class ShapeButton(QtGui.QToolButton):
    def __init__(self, shape, parent=None):
        super(ShapeButton, self).__init__(parent)
        self.shape = shape

        self.vbox = QtGui.QVBoxLayout(self)
        self.vbox.setAlignment(
            QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter
        )  # Align all items centrally
        self.vbox.setContentsMargins(0, 0, 0, 0)
        self.vbox.setSpacing(0)

        self.icon_widget = ShapeWidget(self.shape, QtCore.QSize(71, 70))

        self.label_widget = QtGui.QLabel(shape.label)
        self.label_widget.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)

        self.id_widget = QtGui.QLabel(shape.id)
        self.id_widget.setAlignment(QtCore.Qt.AlignHCenter | QtCore.Qt.AlignVCenter)
        self.id_widget.setContentsMargins(0, 0, 0, 0)
        font = self.id_widget.font()
        font.setPointSize(font.pointSize() * 0.9)
        self.id_widget.setFont(font)

        # Add widgets to the new layout
        self.vbox.addWidget(self.icon_widget)
        self.vbox.addWidget(self.label_widget)
        self.vbox.addWidget(self.id_widget)

        self.setLayout(self.vbox)

        self.setFixedSize(128, 128)
        self.setBaseSize(128, 128)

    def set_text(self, text):
        # Update the text of the label widget
        self.label_widget.setText(text)
